from datetime import datetime

import geopandas as gpd
from shapely.geometry import Point


class TropicalCycloneTrack:
    def __init__(self):
        pass

    def read(self, filename, fmt):
        
            # If ddb_cyc
        if fmt == "ddb_cyc":
            # Read all the lines first
            with open(filename, "r") as f:
                lines = f.readlines()

            # Define the name first
            for line in lines:
                if line[0:4] == "Name":
                    string_value = line[5:]
                    string_value = "".join(ch for ch in string_value if ch.isalnum())
                    self.name = string_value

            # Define other variables names (if they exist)
            for i in range(len(lines)):
                line = lines[i]
                if line[0:11] == "WindProfile":
                    string_value = line[23:]
                    string_value = "".join(ch for ch in string_value if ch.isalnum())
                    self.wind_profile = string_value
                if line[0:20] == "WindPressureRelation":
                    string_value = line[23:]
                    string_value = "".join(ch for ch in string_value if ch.isalnum())
                    self.wind_pressure_relation = string_value
                if line[0:12] == "RMaxRelation":
                    string_value = line[23:]
                    string_value = "".join(ch for ch in string_value if ch.isalnum())
                    self.rmw_relation = string_value
                if line[0:18] == "Backgroundpressure":
                    string_value = line[23:]
                    self.background_pressure = float(string_value)
                if line[0:9] == "PhiSpiral":
                    string_value = line[23:]
                    self.phi_spiral = float(string_value)
                if line[0:20] == "WindConversionFactor":
                    string_value = line[23:]
                    self.wind_conversion_factor = float(string_value)
                if line[0:15] == "SpiderwebRadius":
                    string_value = line[23:]
                    self.spiderweb_radius = float(string_value)
                if line[0:12] == "NrRadialBins":
                    string_value = line[23:]
                    self.nr_radial_bins = int(string_value)
                if line[0:17] == "NrDirectionalBins":
                    string_value = line[23:]
                    self.nr_directional_bins = int(string_value)

            # Read the track
            for i in range(len(lines)):
                line = lines[i]
                if line[0:15] == "##    Datetime " or line[0:15] == "#   Date   Time":
                    break

            # Place coordinates in Tropical Cyclone Track
            for j in range(i + 2, len(lines)):
                # Get values
                line = lines[j]
                line = line.split()
                date_format = "%Y%m%d %H%M%S"
                date_string = line[0] + " " + line[1]
                tc_time = datetime.strptime(date_string, date_format)
                tc_time_string = tc_time.strftime(date_format)
                y = float(line[2])
                x = float(line[3])
                vmax = float(line[4])
                pc = float(line[5])
                RMW = float(line[6])
                R35_NE = float(line[7])
                R35_SE = float(line[8])
                R35_SW = float(line[9])
                R35_NW = float(line[10])
                R50_NE = float(line[11])
                R50_SE = float(line[12])
                R50_SW = float(line[13])
                R50_NW = float(line[14])
                R65_NE = float(line[15])
                R65_SE = float(line[16])
                R65_SW = float(line[17])
                R65_NW = float(line[18])
                R100_NE = float(line[19])
                R100_SE = float(line[20])
                R100_SW = float(line[21])
                R100_NW = float(line[22])

                # Make GeoDataFrame
                point = Point(x, y)
                gdf = gpd.GeoDataFrame(
                    {
                        "datetime": [tc_time_string],
                        "geometry": [point],
                        "vmax": [vmax],
                        "pc": [pc],
                        "RMW": [RMW],
                        "R35_NE": [R35_NE],
                        "R35_SE": [R35_SE],
                        "R35_SW": [R35_SW],
                        "R35_NW": [R35_NW],
                        "R50_NE": [R50_NE],
                        "R50_SE": [R50_SE],
                        "R50_SW": [R50_SW],
                        "R50_NW": [R50_NW],
                        "R65_NE": [R65_NE],
                        "R65_SE": [R65_SE],
                        "R65_SW": [R65_SW],
                        "R65_NW": [R65_NW],
                        "R100_NE": [R100_NE],
                        "R100_SE": [R100_SE],
                        "R100_SW": [R100_SW],
                        "R100_NW": [R100_NW],
                    }
                )
                gdf.set_crs(epsg=self.EPSG, inplace=True)

                # Append self
                self.track = pd.concat([self.track, gdf])

            # Done with this
            self.track = self.track.reset_index(drop=True)
            self.track = self.track.drop([0])  # remove the dummy
            self.track = self.track.reset_index(drop=True)
            if self.debug == 1:
                print("Successfully read track - ddb_cyc")

        elif fmt == "trk":
            # Initialize variables
            lasttime = np.datetime64(datetime.strptime("2000010118", "%Y%m%d%H"))
            newtime2 = lasttime.astype("O")
            tc_time_string = newtime2.strftime("%Y%m%d %H%M%S")
            it = 0
            point = Point(0, 0)
            gdf = gpd.GeoDataFrame(
                {
                    "datetime": [tc_time_string],
                    "geometry": [point],
                    "vmax": [-999],
                    "pc": [-999],
                    "RMW": [-999],
                    "R35_NE": [-999],
                    "R35_SE": [-999],
                    "R35_SW": [-999],
                    "R35_NW": [-999],
                    "R50_NE": [-999],
                    "R50_SE": [-999],
                    "R50_SW": [-999],
                    "R50_NW": [-999],
                    "R65_NE": [-999],
                    "R65_SE": [-999],
                    "R65_SW": [-999],
                    "R65_NW": [-999],
                    "R100_NE": [-999],
                    "R100_SE": [-999],
                    "R100_SW": [-999],
                    "R100_NW": [-999],
                }
            )
            gdf.set_crs(epsg=self.EPSG, inplace=True)
            self.track = gdf  # Always start clean

            # Open the file
            with open(filename, "r") as fid:
                # Read line by line
                for s0 in fid:
                    s0 = s0.strip()
                    if not s0:
                        continue

                    # Start
                    s = s0.split(",")
                    tc = {"name": "not_known"}
                    tc["basin"] = s[0]
                    tc["storm_number"] = int(s[1])

                    # Read time
                    tstr = str(s[2])
                    tstr = tstr.replace(" ", "")
                    time = datetime.strptime(tstr, "%Y%m%d%H")
                    newtime = np.datetime64(time)

                    # Add forecasted time
                    hrs = float(s[5])
                    newtime = newtime + np.timedelta64(int(hrs * 3600), "s")
                    if newtime > lasttime + np.timedelta64(1, "s"):
                        # Update all variables in gdf (when it is not the first one)
                        if it > 0:
                            gdf["geometry"] = [Point(x, y)]  # Assign the new geometry
                            gdf.vmax = vmax
                            gdf.pc = pc
                            gdf.RMW = rmax
                            if R35_NE > 0:
                                gdf.R35_NE = R35_NE
                            if R35_SE > 0:
                                gdf.R35_SE = R35_SE
                            if R35_SW > 0:
                                gdf.R35_SW = R35_SW
                            if R35_NW > 0:
                                gdf.R35_NW = R35_NW

                            if R50_NE > 0:
                                gdf.R50_NE = R50_NE
                            if R50_SE > 0:
                                gdf.R50_SE = R50_SE
                            if R50_SW > 0:
                                gdf.R50_SW = R50_SW
                            if R50_NW > 0:
                                gdf.R50_NW = R50_NW

                            if R65_NE > 0:
                                gdf.R65_NE = R65_NE
                            if R65_SE > 0:
                                gdf.R65_SE = R65_SE
                            if R65_SW > 0:
                                gdf.R65_SW = R65_SW
                            if R65_NW > 0:
                                gdf.R65_NW = R65_NW

                            if R100_NE > 0:
                                gdf.R100_NE = R100_NE
                            if R100_SE > 0:
                                gdf.R100_SE = R100_SE
                            if R100_SW > 0:
                                gdf.R100_SW = R100_SW
                            if R100_NW > 0:
                                gdf.R100_NW = R100_NW

                        # New time point found
                        it += 1
                        lasttime = newtime

                        # Append self
                        self.track = pd.concat([self.track, gdf])

                        # Make TC string
                        newtime2 = newtime.astype("O")
                        tc_time_string = newtime2.strftime("%Y%m%d %H%M%S")

                        # Start fresh
                        gdf = gpd.GeoDataFrame(
                            {
                                "datetime": [tc_time_string],
                                "geometry": [point],
                                "vmax": [-999],
                                "pc": [-999],
                                "RMW": [-999],
                                "R35_NE": [-999],
                                "R35_SE": [-999],
                                "R35_SW": [-999],
                                "R35_NW": [-999],
                                "R50_NE": [-999],
                                "R50_SE": [-999],
                                "R50_SW": [-999],
                                "R50_NW": [-999],
                                "R65_NE": [-999],
                                "R65_SE": [-999],
                                "R65_SW": [-999],
                                "R65_NW": [-999],
                                "R100_NE": [-999],
                                "R100_SE": [-999],
                                "R100_SW": [-999],
                                "R100_NW": [-999],
                            }
                        )
                        gdf.set_crs(epsg=self.EPSG, inplace=True)

                        # Reset all values
                        x = 0
                        y = 0
                        vmax = -999
                        pc = -999
                        RMW = -999
                        R35_NE = -999
                        R35_SE = -999
                        R35_SW = -999
                        R35_NW = -999

                        R50_NE = -999
                        R50_SE = -999
                        R50_SW = -999
                        R50_NW = -999

                        R65_NE = -999
                        R65_SE = -999
                        R65_SW = -999
                        R65_NW = -999

                        R100_NE = -999
                        R100_SE = -999
                        R100_SW = -999
                        R100_NW = -999

                    # Latitude
                    if s[6][-1] == "N":
                        y = 0.1 * float(s[6][:-1])
                    else:
                        y = -0.1 * float(s[6][:-1])
                    # Longitude
                    if s[7][-1] == "E":
                        x = 0.1 * float(s[7][:-1])
                    else:
                        x = -0.1 * float(s[7][:-1])

                    # vmax
                    if float(s[8]) > 0:
                        vmax = float(s[8])

                    # Pressure and raddi
                    if len(s) > 9:
                        # Pressure
                        if float(s[9]) > 0:
                            pc = float(s[9])

                        # Radii
                        r = float(s[11])
                        if r in [34, 35]:
                            R35_NE = float(s[13])
                            R35_SE = float(s[14])
                            R35_SW = float(s[15])
                            R35_NW = float(s[16])
                        elif r == 50:
                            R50_NE = float(s[13])
                            R50_SE = float(s[14])
                            R50_SW = float(s[15])
                            R50_NW = float(s[16])
                        elif r in [64, 65]:
                            R65_NE = float(s[13])
                            R65_SE = float(s[14])
                            R65_SW = float(s[15])
                            R65_NW = float(s[16])
                        elif r == 100:
                            R100_NE = float(s[13])
                            R100_SE = float(s[14])
                            R100_SW = float(s[15])
                            R100_NW = float(s[16])

                        # Other things
                        if len(s) > 17:
                            if s[17]:
                                pressure_last_closed_isobar = float(s[17])
                            if s[18]:
                                radius_last_closed_isobar = float(s[18])
                            if s[19]:
                                try:
                                    if float(s[19]) > 0:
                                        rmax = float(s[19])
                                except ValueError:
                                    print("Error: Unable to convert to float for RMW")
                                    rmax = -999
                            if len(s) >= 28:
                                if s[27]:
                                    name = s[27]

            # Done with this => rest so track looks good
            self.track = self.track.reset_index(drop=True)
            self.track = self.track.drop([0])  # remove the dummy
            self.track = self.track.reset_index(drop=True)
            self.track = self.track.drop([0])  # remove the dummy
            self.track = self.track.reset_index(drop=True)
            if self.debug == 1:
                print("Successfully read track - trk")

        else:
            raise Exception("This file format is not supported as read track!")


    def write(self, filename, fmt):
        # If ddb_cyc
        if fmt == "ddb_cyc":
            # Open file
            with open(filename, "wt") as f:
                # Print header
                f.writelines(
                    "# Tropical Cyclone Toolbox - Coastal Hazards Toolkit - "
                    + self.creation_time.strftime(dateformat_module)
                    + "\n"
                )

                # Print rest
                f.writelines('Name                   "' + self.name + '"\n')
                f.writelines("WindProfile            " + self.wind_profile + "\n")
                f.writelines(
                    "WindPressureRelation   " + self.wind_pressure_relation + "\n"
                )
                f.writelines("RMaxRelation           " + self.rmw_relation + "\n")
                f.writelines(
                    "Backgroundpressure     " + str(self.background_pressure) + "\n"
                )
                f.writelines("PhiSpiral              " + str(self.phi_spiral) + "\n")
                f.writelines(
                    "WindConversionFactor   " + str(self.wind_conversion_factor) + "\n"
                )
                f.writelines(
                    "SpiderwebRadius        " + str(self.spiderweb_radius) + "\n"
                )
                f.writelines(
                    "NrRadialBins           " + str(self.nr_radial_bins) + "\n"
                )
                f.writelines(
                    "NrDirectionalBins      " + str(self.nr_directional_bins) + "\n"
                )
                epsg = self.track.crs.name
                f.writelines("EPSG                   " + epsg + "\n")
                f.writelines(
                    "UnitIntensity          " + str(self.unit_intensity) + "\n"
                )
                f.writelines("UnitWindRadii          " + str(self.unit_radii) + "\n")

                # Print header for the track
                f.writelines("#  \n")
                f.writelines(
                    "#   Date   Time               Lat        Lon         Vmax       Pc          Rmax         R35(NE)      R35(SE)     R35(SW)     R35(NW)     R50(NE)     R50(SE)    R50(SW)    R50(NW)     R65(NE)     R65(SE)     R65(SW)     R65(NW)    R100(NE)    R100(SE)    R100(SW)    R100(NE)  \n"
                )
                f.writelines("#  \n")

                # Print the actual track
                for i in range(len(self.track)):
                    f.writelines(self.track.datetime[i].rjust(20))
                    coords = self.track.geometry[i]
                    f.writelines(str(round(coords.y, 2)).rjust(12))
                    f.writelines(str(round(coords.x, 2)).rjust(12))

                    f.writelines(str(round(self.track.vmax[i], 1)).rjust(12))
                    f.writelines(str(round(self.track.pc[i], 1)).rjust(12))
                    f.writelines(str(round(self.track.RMW[i], 1)).rjust(12))

                    f.writelines(str(round(self.track.R35_NE[i], 1)).rjust(12))
                    f.writelines(str(round(self.track.R35_SE[i], 1)).rjust(12))
                    f.writelines(str(round(self.track.R35_SW[i], 1)).rjust(12))
                    f.writelines(str(round(self.track.R35_NW[i], 1)).rjust(12))

                    f.writelines(str(round(self.track.R50_NE[i], 1)).rjust(12))
                    f.writelines(str(round(self.track.R50_SE[i], 1)).rjust(12))
                    f.writelines(str(round(self.track.R50_SW[i], 1)).rjust(12))
                    f.writelines(str(round(self.track.R50_NW[i], 1)).rjust(12))

                    f.writelines(str(round(self.track.R65_NE[i], 1)).rjust(12))
                    f.writelines(str(round(self.track.R65_SE[i], 1)).rjust(12))
                    f.writelines(str(round(self.track.R65_SW[i], 1)).rjust(12))
                    f.writelines(str(round(self.track.R65_NW[i], 1)).rjust(12))

                    f.writelines(str(round(self.track.R100_NE[i], 1)).rjust(12))
                    f.writelines(str(round(self.track.R100_SE[i], 1)).rjust(12))
                    f.writelines(str(round(self.track.R100_SW[i], 1)).rjust(12))
                    f.writelines(str(round(self.track.R100_NW[i], 1)).rjust(12))

                    f.writelines("\n")

            if self.debug == 1:
                print("Successfully written track - ddb_cyc")
        else:
            print(
                'For other methods of writing the track; please used the "tc.track.to_file" option'
            )



import datetime
from functools import reduce

import geopandas as gpd
import numpy as np
import pandas as pd
import shapely
import xarray as xr

from .tropical_cyclone import TropicalCyclone


class CycloneTrackDatabase:
    """
    CycloneTrackDatabase is a class that reads in a database of tropical cyclone
    tracks and provides methods to filter the database and return a geopandas
    dataframe of the tracks.

    Parameters
    ----------
    name : str
        Name of the database. Currently only "ibtracs" is supported.
    file_name : str
        File name of the database. Currently only "ibtracs" is supported.

    Attributes
    ----------
    name : str
        Name of the database. Currently only "ibtracs" is supported.
    ds : xarray.Dataset
        Dataset of the database.
    lon : numpy.ndarray
        Longitude of the tracks.
    lat : numpy.ndarray
        Latitude of the tracks.
    basin : list
        Basin of the tracks.
    name : list
        Name of the tracks.
    year : numpy.ndarray
        Year of the tracks.
    nstorms : int
        Number of storms in the database.
    ntimes : int
        Number of time steps in the database.
    """

    def __init__(self, name, file_name=None):
        self.name = name
        self.file_name = file_name
        self.ds = None
        self.lon = None
        self.lat = None
        self.basin = None
        self.name = None
        self.year = None
        self.nstorms = None
        self.ntimes = None

        if name == "ibtracs":
            self._read_ibtracs(file_name)
        else:
            pass

    def _read_ibtracs(self, file_name):
        """
        Read in IBTrACS database

        Parameters
        ----------
        file_name : str
            File name of the IBTrACS database.

        Returns
        -------
        None
        """

        # Read in database
        self.ds = xr.open_dataset(file_name)

        # Convert to numpy arrays
        self.lon = self.ds["lon"].values[:]
        self.lat = self.ds["lat"].values[:]
        self.basin = self.ds["basin"].values[:, 0].astype(str).tolist()
        self.name = self.ds["name"].values[:].astype(str).tolist()
        self.year = self.ds["season"].values[:].astype(int)
        self.nstorms = np.shape(self.lon)[0]
        self.ntimes = np.shape(self.lon)[1]

    def get_track(self, index):
        """
        Get a single track from the database.

        Parameters
        ----------
        index : int
            Index of the track in the database.

        Returns
        -------
        tc : TropicalCyclone
            TropicalCyclone object of the track.
        """

        # Create a TropicalCyclone object
        tc = TropicalCyclone(name=self.name[index])

        # Add track
        for it in range(self.ntimes):
            # Check if track is finite
            if not np.isfinite(self.lon[index, it]):
                break

            # Create a shapely point
            point = shapely.Point(self.lon[index, it], self.lat[index, it])

            # Initialize data for geopandas dataframe
            t = self.ds["time"].values[index, it]
            tc_time_string = datetime.datetime.utcfromtimestamp(
                t.item() / 10**9
            ).strftime("%Y%m%d %H%M%S")
            vmax = self.ds["usa_wind"].values[index, it]
            pc = self.ds["usa_pres"].values[index, it]
            RMW = self.ds["usa_rmw"].values[index, it]
            R35_NE = self.ds["usa_r34"].values[index, it, 0]
            R35_SE = self.ds["usa_r34"].values[index, it, 1]
            R35_SW = self.ds["usa_r34"].values[index, it, 2]
            R35_NW = self.ds["usa_r34"].values[index, it, 3]
            R50_NE = self.ds["usa_r50"].values[index, it, 0]
            R50_SE = self.ds["usa_r50"].values[index, it, 1]
            R50_SW = self.ds["usa_r50"].values[index, it, 2]
            R50_NW = self.ds["usa_r50"].values[index, it, 3]
            R65_NE = self.ds["usa_r64"].values[index, it, 0]
            R65_SE = self.ds["usa_r64"].values[index, it, 1]
            R65_SW = self.ds["usa_r64"].values[index, it, 2]
            R65_NW = self.ds["usa_r64"].values[index, it, 3]
            # R100_NE = self.ds["usa_r100"].values[index, it, 0]
            # R100_SE = self.ds["usa_r100"].values[index, it, 1]
            # R100_SW = self.ds["usa_r100"].values[index, it, 2]
            # R100_NW = self.ds["usa_r100"].values[index, it, 3]
            vmax = np.nan_to_num(vmax, copy=True, nan=-999.0)
            pc = np.nan_to_num(pc, copy=True, nan=-999.0)
            RMW = np.nan_to_num(RMW, copy=True, nan=-999.0)
            R35_NE = np.nan_to_num(R35_NE, copy=True, nan=-999.0)
            R35_SE = np.nan_to_num(R35_SE, copy=True, nan=-999.0)
            R35_SW = np.nan_to_num(R35_SW, copy=True, nan=-999.0)
            R35_NW = np.nan_to_num(R35_NW, copy=True, nan=-999.0)
            R50_NE = np.nan_to_num(R50_NE, copy=True, nan=-999.0)
            R50_SE = np.nan_to_num(R50_SE, copy=True, nan=-999.0)
            R50_SW = np.nan_to_num(R50_SW, copy=True, nan=-999.0)
            R50_NW = np.nan_to_num(R50_NW, copy=True, nan=-999.0)
            R65_NE = np.nan_to_num(R65_NE, copy=True, nan=-999.0)
            R65_SE = np.nan_to_num(R65_SE, copy=True, nan=-999.0)
            R65_SW = np.nan_to_num(R65_SW, copy=True, nan=-999.0)
            R65_NW = np.nan_to_num(R65_NW, copy=True, nan=-999.0)
            R100_NE = -999.0
            R100_SE = -999.0
            R100_SW = -999.0
            R100_NW = -999.0

            # Create a geopandas dataframe
            gdf = gpd.GeoDataFrame(
                {
                    "datetime": [tc_time_string],
                    "geometry": [point],
                    "vmax": [vmax],
                    "pc": [pc],
                    "RMW": [RMW],
                    "R35_NE": [R35_NE],
                    "R35_SE": [R35_SE],
                    "R35_SW": [R35_SW],
                    "R35_NW": [R35_NW],
                    "R50_NE": [R50_NE],
                    "R50_SE": [R50_SE],
                    "R50_SW": [R50_SW],
                    "R50_NW": [R50_NW],
                    "R65_NE": [R65_NE],
                    "R65_SE": [R65_SE],
                    "R65_SW": [R65_SW],
                    "R65_NW": [R65_NW],
                    "R100_NE": [R100_NE],
                    "R100_SE": [R100_SE],
                    "R100_SW": [R100_SW],
                    "R100_NW": [R100_NW],
                }
            )

            # Set CRS coordinate system
            gdf.set_crs(epsg=4326, inplace=True)

            # Append self
            tc.track = pd.concat([tc.track, gdf])

        # Done with this
        tc.track = tc.track.reset_index(drop=True)
        tc.track = tc.track.drop([0])  # remove the dummy
        tc.track = tc.track.reset_index(drop=True)

        return tc

    def list_names(self, index=None):
        if index is not None:
            return [self.name[i] for i in index]
        else:
            return self.name

    def to_gdf(self, index=None):
        """
        Returns a geopandas dataframe of the tracks.

        Parameters
        ----------
        index : list
            List of indices of the tracks to return. If None, all tracks are returned.

        Returns
        -------
        gdf : geopandas.GeoDataFrame
            Geopandas dataframe of the tracks.
        """

        # Initialize
        geom = []
        iok = []
        if index is None:
            index = range(self.nstorms)
        description = []

        # Loop over tracks
        for ind in index:
            lon = self.lon[ind, :]
            lat = self.lat[ind, :]
            lon = lon[np.isfinite(lon)]
            lat = lat[np.isfinite(lat)]
            if len(lon) > 1 and len(lat) > 1:
                geom.append(
                    shapely.geometry.LineString(np.transpose(np.stack((lon, lat))))
                )
                iok.append(ind)
                description.append(self.name[ind] + " (" + str(self.year[ind]) + ")")

        # Create geopandas dataframe
        gdf = gpd.GeoDataFrame(crs=4326, geometry=geom)
        names = [self.name[i] for i in iok]
        years = self.year[iok]
        gdf["name"] = names
        gdf["year"] = years
        gdf["description"] = description
        gdf["database_index"] = iok
        return gdf

    def filter(
        self,
        name=None,
        distance=None,
        lon=None,
        lat=None,
        basin=None,
        year=None,
        year_min=None,
        year_max=None,
        vmax_min=None,
        vmax_max=None,
    ):
        """
        Filter the database and return the indices of the filtered tracks.

        Parameters
        ----------
        name : str
            Name of the track to filter by.
        distance : float
            Distance (km) of the track to the point (lon, lat).
        lon : list
            Longitude range of the tracks.
        lat : list
            Latitude range of the tracks.
        basin : str
            Basin of the tracks.
        year : int
            Year of the tracks.
        year_min : int
            Minimum year of the tracks.
        year_max : int
            Maximum year of the tracks.

        Returns
        -------
        index : list
            List of indices of the filtered tracks.
        """

        # Initialize year_min and year_max
        if year:
            year_min = year
            year_max = year
        else:
            if not year_min:
                year_min = 0
            if not year_max:
                year_max = 9999

        if not vmax_min:
            vmax_min = 0.0
        if not vmax_max:
            vmax_max = 9999.0

        # Filter by basin
        if basin:
            ibasin = np.array(
                [
                    i
                    for i in range(len(self.basins))
                    if self.basin[i].lower() == basin.lower()
                ]
            )
        else:
            ibasin = np.arange(0, self.nstorms)

        # Filter by year
        if year_min and year_max:
            iyear = np.where((self.year >= year_min) & (self.year <= year_max))[0]
        else:
            iyear = np.arange(0, self.nstorms)

        # # Filter by vmax
        # if vmax_min and vmax_max:
        #     ivmax = np.where((self.year >= year_min) & (self.year <= year_max))[0]
        # else:
        #     iyear = np.arange(0, self.nstorms)

        # Filter by name
        if name:
            iname = np.array(
                [
                    i
                    for i in range(len(self.name))
                    if self.name[i].lower() == name.lower()
                ]
            )
        else:
            iname = np.arange(0, self.nstorms)

        # Filter by bounding box
        if isinstance(lon, list) and isinstance(lat, list):
            inear = np.where(
                (self.lon >= lon[0])
                & (self.lon <= lon[1])
                & (self.lat >= lat[0])
                & (self.lat <= lat[1])
            )
            ibbox = np.unique(inear[0])
        else:
            ibbox = np.arange(0, self.nstorms)

        # Filter by distance
        if distance:
            # Compute distance of all tracks to point
            d = CycloneTrackDatabase.compute_distance(lon, lat, self.lon, self.lat)
            dmin = np.nanmin(d, axis=1)
            idist = np.where(dmin < distance)[0]
        else:
            idist = np.arange(0, self.nstorms)

        # Intersect all filters
        index = reduce(np.intersect1d, (ibasin, iyear, iname, idist, ibbox))

        return index

    @staticmethod
    def compute_distance(lon1, lat1, lon2, lat2):
        """
        Compute the distance between two points on a sphere.

        Parameters
        ----------
        lon1 : numpy.ndarray
            Longitude of the first point.
        lat1 : numpy.ndarray
            Latitude of the first point.
        lon2 : numpy.ndarray
            Longitude of the second point.
        lat2 : numpy.ndarray
            Latitude of the second point.

        Returns
        -------
        d : numpy.ndarray
            Distance between the points.
        """

        # Convert to radians
        R = 6373.0
        lon1 = lon1 * np.pi / 180
        lat1 = lat1 * np.pi / 180
        lon2 = lon2 * np.pi / 180
        lat2 = lat2 * np.pi / 180
        dlon = lon2 - lon1

        # Wrap around
        dlon[np.where(dlon < -np.pi)] = dlon[np.where(dlon < -np.pi)] + 2 * np.pi
        dlon[np.where(dlon > np.pi)] = dlon[np.where(dlon > np.pi)] - 2 * np.pi
        dlat = lat2 - lat1

        # Compute distance
        a = (np.sin(dlat / 2)) ** 2 + np.cos(lat1) * np.cos(lat2) * (
            np.sin(dlon / 2)
        ) ** 2
        c = 2 * np.arctan2(np.sqrt(a), np.sqrt(1.0 - a))
        return R * c



# -*- coding: utf-8 -*-
"""
Created on Sun Apr 25 10:58:08 2021

@author: ormondt
"""

__version__ = "1.0.2"

from .tropical_cyclone_refactored import TropicalCyclone
from .track_database import CycloneTrackDatabase
from .track_dataset import CycloneTrackDataset
from .track_selector import track_selector



import os


def track_selector(dataset, app, lon=0.0, lat=0.0, distance=1000.0, year_min=1850, year_max=2030):

    app.gui.setvar("cyclone_track_selector", "lon", lon)
    app.gui.setvar("cyclone_track_selector", "lat", lat)
    app.gui.setvar("cyclone_track_selector", "distance", distance)
    app.gui.setvar("cyclone_track_selector", "year0", year_min)
    app.gui.setvar("cyclone_track_selector", "year1", year_max)
    app.gui.setvar("cyclone_track_selector", "name", "")

    data = {}
    data["track_dataset"] = dataset

    # Read GUI config file
    config_file = os.path.join(os.path.dirname(__file__), "cyclone_track_selector.yml")
    okay, data = app.gui.popup(config_file, id="track_selector", data=data)

    track = None
    if okay:
        # Get the track from the dataset
        track = dataset.get_track(data["dataset_index"])

    return track, okay

def map_ready(widget):

    print("Selector map is ready !")

    gui = widget.element.gui

    mp = gui.popup_window["track_selector"].find_element_by_id("track_selector_map").widget
    mp.jump_to(0.0, 0.0, 1)
    data = gui.popup_data
    # Container layers
    data["track_selector"]["main_layer"] = mp.add_layer("track_selector")
    # Tracks layers
    data["track_selector"]["track_layer"] = data["track_selector"]["main_layer"].add_layer(
        "tracks",
        type="line_selector",
        file_name="tracks.geojson",
        select=select_track,
        selection_type="single",
        line_color="dodgerblue",
        line_width=2,
        line_color_selected="red",
        line_width_selected=3,
        hover_param="description",
    )

    # Update data in tracks layer
    update_tracks(gui)


def update_tracks(gui):

    data = gui.popup_data

    tdb = data["track_selector"]["track_dataset"]
    tracks_layer = data["track_selector"]["track_layer"]

    # Get filter data
    distance = gui.getvar("cyclone_track_selector", "distance")
    year_min = gui.getvar("cyclone_track_selector", "year0")
    year_max = gui.getvar("cyclone_track_selector", "year1")
    lon      = gui.getvar("cyclone_track_selector", "lon")
    lat      = gui.getvar("cyclone_track_selector", "lat")

    # Get indices based on filter
    index = tdb.filter(
        lon=lon, lat=lat, distance=distance, year_min=year_min, year_max=year_max
    )

    # Get GeoDataFrame of tracks
    gdf = tdb.to_gdf(index=index)

    tracks_layer.set_data(gdf, 0)


def map_moved(coords, widget):
    pass


def select_track(feature, widget):
    widget.element.gui.popup_data["track_selector"]["dataset_index"] = feature["properties"]["dataset_index"]


def edit_filter(val, widget):
    update_tracks(widget.element.gui)



import numpy as np


######
# Definitions that I want to be available in general
######
# Definitions to compute Holland 2010 (1D)
def holland2010(r, vmax, pc, pn, rmax, dpdt, lat, vt, xn):
    """
    Returning the one-dimensional Holland et al. (2010) parametric wind profile

    Parameters
    ----------
    r : radius in km
    etc etc.

    """
    # calculate Holland b parameter based on Holland (2008) - assume Dvorak method
    vms = vmax
    dp = max(pn - pc, 1)
    x = 0.6 * (1 - dp / 215)
    if np.isnan(dpdt):
        dpdt = 0
    b = (
        -4.4 * 10**-5 * dp**2
        + 0.01 * dp
        + 0.03 * dpdt
        - 0.014 * abs(lat)
        + 0.15 * vt**x
        + 1
    )
    b = min(np.nanmax(np.append(b, 1)), 2)  # numerical limits

    # initialise
    x = np.zeros(np.size(r)) + xn
    pr = np.zeros((np.size(r)))
    vr = np.zeros((np.size(r)))

    # Compute
    index = r <= rmax
    x[index] = 0.5
    rp = (rmax / r) ** b
    pr = pc + dp * np.exp(-rp)
    vr = vms * (rp * np.exp(1.0 - rp)) ** x

    # output
    return [vr, pr]


# Definitiaton for wind radii
def wind_radii_nederhoff(vmax, lat, region, probability):
    """
    Returning the estimates of radius of maximum winds (RMW)
    and estimates of gale force winds (R35)

    Parameters
    ----------
    vmax    : wind speed intensity in m/s - 1-minute average
    lat     : latitude in degrees
    region  : see paper for details (use 7 for non region specific)
    probability: integer - 0 only mode; 1 = 1000 values

    Returns
    -------
    rmax    : dictionary with mode and possible more
    r35     : ''
    """
    # radius of maximum winds (rmw or rmax)
    # 1. coefficients for A
    coefficients_a = np.array(
        [
            0.306982540000000,
            0.338409237000000,
            0.342791450000000,
            0.363546490000000,
            0.358572938000000,
            0.310729085000000,
            0.395431764000000,
            0.370190027000000,
        ]
    )

    # 2. coefficients for B
    coefficients_b = np.array(
        [
            [
                132.411906200000,
                14.5640379700000,
                -0.00259703300000000,
                20.3808036500000,
            ],
            [229.245844100000, 9.53865069100000, 0.00398810500000000, 28.4457367200000],
            [85.2576655100000, 30.6920872600000, 0.00243248000000000, 5.78116540600000],
            [127.833300700000, 11.8474757400000, 0.0159363120000000, 25.4682000500000],
            [153.733294700000, 11.4788885400000, 0.00747119300000000, 28.9489788700000],
            [261.528874200000, 7.01151785400000, 0.0261912560000000, 29.2022787100000],
            [19.0899242800000, 24.0885573100000, 0.106240340000000, 23.1802014600000],
            [44.8241743300000, 23.3717128800000, 0.0304690570000000, 22.4282036100000],
        ]
    )

    # 3. get the best guess for a and b given wind speed and latitude
    a_value = coefficients_a[region]
    b_value = (
        coefficients_b[region, 0]
        * np.exp(-vmax / coefficients_b[region, 1])
        * (1 + coefficients_b[region, 2] * abs(lat))
        + coefficients_b[region, 3]
    )

    rmax = {}
    rmax["mode"] = {}
    rmax["mean"] = {}
    rmax["median"] = {}
    rmax["lowest"] = {}
    rmax["highest"] = {}
    rmax["numbers"] = {}

    # 4. compute 1000 delta r35 values
    rmax["mode"] = np.exp(np.log(b_value) - a_value**2)
    if probability == 1:
        numbers = np.sort(
            np.exp(np.random.normal(size=(1000, 1)) * a_value + np.log(b_value))
        )
        rmax["mean"] = np.mean(numbers)
        rmax["median"] = np.median(numbers)
        rmax["lowest"] = numbers[int(0.05 * len(numbers))][0]
        rmax["highest"] = numbers[int(0.95 * len(numbers))][0]
        rmax["numbers"] = np.sort(numbers)

    # delta radius of 35 knots (r35)
    dr35 = {}
    dr35["mode"] = {}
    dr35["mean"] = {}
    dr35["median"] = {}
    dr35["lowest"] = {}
    dr35["highest"] = {}
    dr35["numbers"] = {}

    # Only if wind speed is more than 20 m/s
    if vmax > 20:
        # 1. coefficients for a
        coefficients_a = np.array(
            [
                [0.121563729, -0.052184289, 0.032953813],
                [0.131188105, -0.044389473, 0.002253258],
                [0.122286754, -0.045355772, 0.013286154],
                [0.120490659, -0.035029431, -0.005249445],
                [0.156059522, -0.041685377, 0.004952978],
                [-0.251333213, -0.009072243, -0.00506365],
                [0.131903526, -0.042096876, 0.012443195],
                [0.190044585, -0.044602083, 0.006117124],
            ]
        )

        # 2. coefficients for b
        coefficients_b = np.array(
            [
                [30.92867473, 0.530681714, -0.012001645],
                [30.21210133, 0.414897465, 0.021689596],
                [26.58686237, 0.425916004, 0.028547278],
                [23.88007085, 0.43109144, 0.038119083],
                [33.26829485, 0.42859578, 0.017209431],
                [18.11013691, 0.486399912, 0.02955688],
                [16.9973011, 0.453713419, 0.054643743],
                [29.61141102, 0.4132484, 0.024418947],
            ]
        )

        # 3. get the best guess for a and b given wind speed and latitude
        a_value = coefficients_a[region, 0] + np.exp(
            vmax * coefficients_a[region, 1]
        ) * (1 + coefficients_a[region, 2] * abs(lat))
        b_value = (
            coefficients_b[region, 0]
            * (vmax - 18) ** coefficients_b[region, 1]
            * (1 + coefficients_b[region, 2] * abs(lat))
        )

        # 4. compute 1000 delta r35 values
        dr35["mode"] = np.exp(np.log(b_value) - a_value**2)
        if probability == 1:
            numbers = np.sort(
                np.exp(np.random.normal(size=(1000, 1)) * a_value + np.log(b_value))
            )
            dr35["mean"] = np.mean(numbers)
            dr35["median"] = np.median(numbers)
            dr35["lowest"] = numbers[int(0.05 * len(numbers))][0]
            dr35["highest"] = numbers[int(0.95 * len(numbers))][0]
            dr35["numbers"] = np.sort(numbers)

    # output
    return [rmax, dr35]


# Definition to compute wind-pressure relation to determine the vmax or the pressure drop
def wpr_holland2008(
    pc=None, pn=None, phi=None, vt=None, dpcdt=None, rhoa=None, SST=None, vmax=None
):
    # used when pc needs to be determined
    if not rhoa:
        if vmax:
            dp1 = np.arange(1, 151 + 5, 5)
            pc1 = pn - dp1
        else:
            dp1 = pn - pc
            pc1 = pc

        if not SST:
            Ts = 28.0 - 3 * (phi - 10) / 20  # surface temperature
        else:
            Ts = SST - 1

        prmw = pc1 + dp1 / 3.7
        qm = 0.9 * (3.802 / prmw) * np.exp(17.67 * Ts / (243.5 + Ts))  # vapor pressure
        Tvs = (Ts + 273.15) * (1.0 + 0.81 * qm)  # virtual surface air temperature
        Rspecific = 287.058
        rhoa = 100 * pc1 / (Rspecific * Tvs)

    # vmax to be determined
    if not vmax:
        pc = min(pc, pn - 1.0)
        dp = pn - pc
        x = 0.6 * (1 - dp / 215)
        bs = (
            -4.4e-5 * dp**2
            + 0.01 * dp
            + 0.03 * dpcdt
            - 0.014 * phi
            + 0.15 * vt**x
            + 1.0
        )
        vmax = np.sqrt(100 * bs * dp / (rhoa * np.e))
        output = vmax
    else:
        vtkmh = vt * 3.6
        vmaxkmh = vmax * 3.6 * 0.88
        dp = (
            0.00592
            * (1 - 0.0687 * vtkmh**0.33)
            * (1 + 0.00285 * abs(phi) ** 1.35)
            * vmaxkmh**1.81
        )
        output = pn - dp

    return output







# Load modules
import os

from cht_cyclones.tropical_cyclone import TropicalCyclone, TropicalCycloneEnsemble

# Define a track
tc                  = TropicalCyclone(name="Idai")

# Read a track
current_directory   = os.path.dirname(os.path.abspath(__file__))
file_path           = os.path.join(current_directory, 'best_track_idai.cyc')
tc.from_ddb_cyc(file_path)

# Define settings
# Uses typical Wind Enhance Scheme (WES) formulations
tc.include_rainfall     = True      # using empirical formulation to compute rainfall (IPET default)
tc.rainfall_factor      = 2.0       # factor to calibrate rainfall 2.0 means 2x as much as relationship

# create (regular) ASCII spiderweb
file_path           = os.path.join(current_directory, 'best_track_idai.spw')
tc.to_spiderweb(file_path)

# create (netcdf) spiderweb
file_path           = os.path.join(current_directory, 'best_track_idai.nc')
tc.to_spiderweb(file_path, format_type='netcdf')

# Write out as shapefile
file_path           = os.path.join(current_directory, 'best_track_idai.shp')
tc.track.to_file(file_path)

# Write out as geojson
file_path           = os.path.join(current_directory, 'best_track_idai.json')
tc.track.to_file(file_path, driver="GeoJSON")

# Write as cyc again
tc.convert_units_metric_imperial()
file_path           = os.path.join(current_directory, 'best_track_idai_cht.cyc')
tc.write_track(file_path, 'ddb_cyc')


import numpy as np
from scipy.interpolate import CubicSpline, interp1d

from .wind_profiles import holland2010


# Definition to fit Holland 2010 wind field
def fit_wind_field_holland2010(
    vmax, rmax, pc, vtreal, phit, pn, phi_spiral, lat, dpdt, obs
):
    # Discussion
    # shouldnt we have a switch to only calibrate vt and phi_a for observed radii
    # what about limits on these variables
    # OK with xn calibration

    # function to fit wind field based on Holland 2010
    size_factor = 1
    phi = np.arange(90, -270 - 10, -10)  # radial angles (cartesian, degrees)
    rmax = rmax * 1000  # convert from km to m
    r = np.arange(5000, 500000 + 5000, 5000)

    # first estimates
    xn = 0.5
    vt = 0.6 * vtreal

    if lat > 0:
        phia = 45  # angle with respect to track angle (cartesian degrees, i.e. counter-clockwise)
    else:
        phia = -45

    # More variables
    dxn = 0.01
    dvt = 0.5
    dphia = 5
    nrad = 2
    nobs = 0

    for irad in range(np.size(obs["quadrants_radii"], 0)):
        for iquad in range(np.size(obs["quadrants_radii"], 1)):
            if not np.isnan(obs["quadrants_radii"][irad, iquad]):
                nobs = nobs + 1
    wrad = obs["quadrants_speed"]            

    # just for plotting
    xx = np.zeros((len(phi), len(r)))
    yy = np.zeros((len(phi), len(r)))
    # plt.pcolor(xx,yy,w)

    for j in range(len(phi)):
        for h in range(len(r)):
            xx[j, h] = 0.001 * r[h] * np.cos(phi[j] * np.pi / 180)
            yy[j, h] = 0.001 * r[h] * np.sin(phi[j] * np.pi / 180)

    if nobs > 0 and vmax > 21.0:
        # Do three fits
        for irep in range(3):
            # Default fit
            w = compute_wind_field(
                r,
                phi,
                vmax,
                pc,
                rmax,
                pn,
                vtreal,
                phit,
                lat,
                dpdt,
                phi_spiral,
                xn,
                vt,
                phia,
            )
            [mean_error, rms_error, err] = compute_mean_error(r, w, obs, wrad)
            nit = 0

            # 1) now first adjust size of storm with xn
            while True:
                nit = nit + 1
                wu = compute_wind_field(
                    r,
                    phi,
                    vmax,
                    pc,
                    rmax,
                    pn,
                    vtreal,
                    phit,
                    lat,
                    dpdt,
                    phi_spiral,
                    xn + dxn,
                    vt,
                    phia,
                )
                [mean_error_u, rms_error_u, err_u] = compute_mean_error(
                    r, wu, obs, wrad
                )
                wd = compute_wind_field(
                    r,
                    phi,
                    vmax,
                    pc,
                    rmax,
                    pn,
                    vtreal,
                    phit,
                    lat,
                    dpdt,
                    phi_spiral,
                    xn - dxn,
                    vt,
                    phia,
                )
                [mean_error_d, rms_error_d, err_d] = compute_mean_error(
                    r, wd, obs, wrad
                )

                # better with larger value of xn
                if rms_error_u < rms_error:
                    xn = xn + dxn
                    rms_error = rms_error_u

                # better with smaller value of xn
                elif rms_error_d < rms_error:
                    xn = xn - dxn
                    rms_error = rms_error_d

                # optimum reached
                else:
                    break

                # Other criteria
                if xn < 0.3 or xn > 1.0:
                    break
                if nit > 100:
                    break

            nit = 0

            # 2) now the asymmetry magnitude
            while True:
                nit = nit + 1
                wu = compute_wind_field(
                    r,
                    phi,
                    vmax,
                    pc,
                    rmax,
                    pn,
                    vtreal,
                    phit,
                    lat,
                    dpdt,
                    phi_spiral,
                    xn,
                    vt + dvt,
                    phia,
                )
                [mean_error_u, rms_error_u, err_u] = compute_mean_error(
                    r, wu, obs, wrad
                )
                wd = compute_wind_field(
                    r,
                    phi,
                    vmax,
                    pc,
                    rmax,
                    pn,
                    vtreal,
                    phit,
                    lat,
                    dpdt,
                    phi_spiral,
                    xn,
                    vt - dvt,
                    phia,
                )
                [mean_error_d, rms_error_d, err_d] = compute_mean_error(
                    r, wd, obs, wrad
                )

                # better with larger value of vt
                if rms_error_u < rms_error:
                    vt = vt + dvt
                    rms_error = rms_error_u

                # better with smaller value of vt
                elif rms_error_d < rms_error:
                    vt = vt - dvt
                    rms_error = rms_error_d

                # optimum reached
                else:
                    break

                # Other criteria
                if vt < 0.0 or vt > 1.5 * vtreal:
                    break
                if nit > 100:
                    break

            nit = 0

            # 3. now the asymmetry direction
            while True:
                nit = nit + 1

                wu = compute_wind_field(
                    r,
                    phi,
                    vmax,
                    pc,
                    rmax,
                    pn,
                    vtreal,
                    phit,
                    lat,
                    dpdt,
                    phi_spiral,
                    xn,
                    vt,
                    phia + dphia,
                )
                [mean_error_u, rms_error_u, err_u] = compute_mean_error(
                    r, wu, obs, wrad
                )
                wd = compute_wind_field(
                    r,
                    phi,
                    vmax,
                    pc,
                    rmax,
                    pn,
                    vtreal,
                    phit,
                    lat,
                    dpdt,
                    phi_spiral,
                    xn,
                    vt,
                    phia - dphia,
                )
                [mean_error_d, rms_error_d, err_d] = compute_mean_error(
                    r, wd, obs, wrad
                )

                if rms_error_u < rms_error:
                    # better with larger value of phia
                    phia = phia + dphia
                    rms_error = rms_error_u

                elif rms_error_d < rms_error:
                    # better with smaller value of phia
                    phia = phia - dphia
                    rms_error = rms_error_d
                else:
                    # optimum reached
                    break

                if phia < -180:
                    phia = phia + 360

                if phia > 180:
                    phia = phia - 360

                if nit > 100:
                    break

    return [xn, vt, phia]


# Definition to compute mean error
def compute_mean_error(r, w, obs, wrad):
    # Discussion
    # why are we only accounting for R35?

    # variables
    nrad = np.size(obs["quadrants_radii"], 0)
    nrad = 3  # not we are only fitting R35, nothing else
    nq = np.size(obs["quadrants_radii"], 1)
    iq1 = [0, 9, 18, 27]
    iq2 = [9, 18, 27, 36]
    err = np.zeros((nq, nrad))

    # Go over the quadrants and radii
    for irad in range(nrad):
        for iquad in range(nq):
            vrad = 0
            for j in range(iq1[iquad], iq2[iquad] + 1):
                ww = w[j, :]
                if not np.isnan(obs["quadrants_radii"][irad, iquad]):
                    # compute wind speed vrad at required radius
                    wf = interp1d(r, ww, bounds_error=False, fill_value=0)
                    w0 = wf(obs["quadrants_radii"][irad, iquad] * 1000)
                    vrad = max(vrad, w0)
                else:
                    # maximum wind speed must be lower than wrad
                    w0 = max(ww)
                    vrad = max(vrad, w0)
            if not np.isnan(obs["quadrants_radii"][irad, iquad]):
                err[iquad, irad] = vrad - wrad[irad]
            else:
                err[iquad, irad] = np.nan

    # Get error values
    mask = ~np.isnan(err)  # Create the mask
    err = err[mask]
    mean_error = np.nanmean(err)
    rms_error = np.sqrt(np.mean(err**2))

    # Return
    return [mean_error, rms_error, err]


# Definition to compute forward speed and heading
def compute_forward_speed_heading(t, x, y):
    # variables
    forward_speed = np.zeros((len(x)))
    heading = np.zeros((len(x)))
    for it in range(len(x)):
        # Get basics
        geofacy = 111111
        geofacx = geofacy * np.cos(y[it] * np.pi / 180)

        if it == 0:
            # print('Forward')
            # datetime_forward    =
            # dt                  = datetime_forward - datetime_it
            # dt                  = dt.total_seconds()
            dx = (x[it + 1] - x[it]) * geofacx
            dy = (y[it + 1] - y[it]) * geofacy

        elif it == (len(x) - 1):
            # print('Backward')
            # datetime_backward   = datetime.strptime(self.track.datetime[it-1], dateformat_module)
            # coords_backward     = self.track.geometry[it-1]
            # dt                  = datetime_it - datetime_backward
            # dt                  = dt.total_seconds()
            dx = (x[it] - x[it - 1]) * geofacx
            dy = (y[it] - y[it - 1]) * geofacy

        else:
            # Backward
            dx1 = (x[it] - x[it - 1]) * geofacx
            dy1 = (y[it] - y[it - 1]) * geofacy

            # Forward
            dx2 = (x[it + 1] - x[it]) * geofacx
            dy2 = (y[it + 1] - y[it]) * geofacy

            # Combined yields central differences
            # print('Central')
            dx = np.mean([dx1, dx2])
            dy = np.mean([dy1, dy2])

        # Compute angle
        forward_speed[it] = 0.0
        heading[it] = np.arctan2(dy, dx)

    # Return
    return [forward_speed, heading]


# definition to compute wind field
def compute_wind_field(
    r, phi, vmax, pc, rmax, pn, vtreal, phit, lat, dpdt, phi_spiral, xn, vt, phia
):
    # Discussion is asymmetry account for properly? I believe there should be a factor in front of ux/vy
    vms = vmax - vt

    # compute wind profile (vr and pr)
    [vr, pr] = holland2010(r, vms, pc, pn, rmax, dpdt, lat, vtreal, xn)

    wind_speed = np.zeros((phi.shape[0], r.shape[0]))
    wind_to_direction_cart = np.zeros((phi.shape[0], r.shape[0]))
    for iphi in range(len(phi)):
        wind_speed[iphi, :] = vr
        if lat >= 0:
            # northern hemisphere
            dr = 90 + phi[iphi] + phi_spiral
        else:
            # southern hemisphere
            dr = -90 + phi[iphi] - phi_spiral
        wind_to_direction_cart[iphi, :] = dr

    vnorm = wind_speed / np.max(wind_speed)
    vnorm = np.zeros(np.shape(wind_speed)) + 1.0

    ux = vt * np.cos((phit + phia) * np.pi / 180)
    uy = vt * np.sin((phit + phia) * np.pi / 180)

    vx = wind_speed * np.cos(wind_to_direction_cart * np.pi / 180) + ux * vnorm
    vy = wind_speed * np.sin(wind_to_direction_cart * np.pi / 180) + uy * vnorm

    wind_speed = np.sqrt(vx**2 + vy**2)

    return wind_speed



import os
from pathlib import Path

from cht_cyclones.cyclone_track_database import CycloneTrackDatabase


def test_load_track_database():
    database_file = Path(os.getcwd()).joinpath(
        "cyclone_track_database", "IBTrACS.NA.v04r00.nc"
    )
    tc = CycloneTrackDatabase("ibtracs", file_name=database_file)

    assert tc.nstorms == 2323



import datetime
import os
from functools import reduce

import boto3
import geopandas as gpd
import numpy as np
import pandas as pd
import shapely
import toml
import xarray as xr
from botocore import UNSIGNED
from botocore.client import Config
from cht_utils import fileops as fo

from .tropical_cyclone_refactored import TropicalCyclone


class CycloneTrackDataset:
    """
    CycloneTrackDataset is a class that reads in a dataset of tropical cyclone
    tracks and provides methods to filter the dataset and return a geopandas
    dataframe of the tracks.

    Parameters
    ----------
    name : str
        Name of the dataset. Currently only "ibtracs" is supported.
    file_name : str
        File name of the dataset. Currently only "ibtracs" is supported.

    Attributes
    ----------
    name : str
        Name of the dataset. Currently only "ibtracs" is supported.
    ds : xarray.Dataset
        Dataset of the dataset.
    lon : numpy.ndarray
        Longitude of the tracks.
    lat : numpy.ndarray
        Latitude of the tracks.
    basin : list
        Basin of the tracks.
    name : list
        Name of the tracks.
    year : numpy.ndarray
        Year of the tracks.
    nstorms : int
        Number of storms in the dataset.
    ntimes : int
        Number of time steps in the dataset.
    """

    def __init__(self, name, path):
        self.name = name
        self.path = path
        self.ds = None
        self.lon = None
        self.lat = None
        self.basin = None
        self.storm_name = None
        self.year = None
        self.nstorms = None
        self.ntimes = None

        self.read_metadata()

    def read(self):
        if self.ds is not None:
            # Already read in
            return
        filename = os.path.join(self.path, self.files[0])
        self._read_ibtracs(filename)

    def read_metadata(self):
        # Read metadata file
        tml_file = os.path.join(self.path, "metadata.tml")
        tml = toml.load(tml_file)
        for key in tml:
            setattr(self, key, tml[key])
        # Long name for backwards compatibility
        if "longname" in tml:
            self.long_name = tml["longname"]
        # Make sure there is always a long_name
        if self.long_name == "":
            self.long_name = self.name        

    def download(self):
        if self.s3_bucket is None:
            return
        # Check if download is needed
        for file in self.files:
            if not os.path.exists(os.path.join(self.path, file)):
                s3_client = boto3.client('s3', config=Config(signature_version=UNSIGNED))
                break
        # Get all files defined in the toml file
        for file in self.files:
            if not os.path.exists(os.path.join(self.path, file)):
                print(f"Downloading {file} from track dataset {self.name} ...")
                s3_client.download_file(self.s3_bucket, f"{self.s3_key}/{file}", os.path.join(self.path, file))

    def _read_ibtracs(self, file_name):
        """
        Read in IBTrACS dataset

        Parameters
        ----------
        file_name : str
            File name of the IBTrACS dataset.

        Returns
        -------
        None
        """

        # Read in dataset
        self.ds = xr.open_dataset(file_name)

        # Convert to numpy arrays
        self.lon = self.ds["lon"].values[:]
        self.lat = self.ds["lat"].values[:]
        self.basin = self.ds["basin"].values[:, 0].astype(str).tolist()
        self.storm_name = self.ds["name"].values[:].astype(str).tolist()
        self.year = self.ds["season"].values[:].astype(int)
        self.nstorms = np.shape(self.lon)[0]
        self.ntimes = np.shape(self.lon)[1]

    def get_track(self, index):
        """
        Get a single track from the dataset.

        Parameters
        ----------
        index : int
            Index of the track in the dataset.

        Returns
        -------
        tc : TropicalCyclone
            TropicalCyclone object of the track.
        """

        # Create a TropicalCyclone object
        tc = TropicalCyclone(name=self.storm_name[index])

        gdf = gpd.GeoDataFrame()

        # Add track
        for it in range(self.ntimes):
            # Check if track is finite
            if not np.isfinite(self.lon[index, it]):
                break

            # Create a shapely point
            point = shapely.Point(self.lon[index, it], self.lat[index, it])

            # Initialize data for geopandas dataframe
            t = self.ds["time"].values[index, it]
            tc_time_string = datetime.datetime.utcfromtimestamp(
                t.item() / 10**9
            ).strftime("%Y%m%d %H%M%S")
            vmax = self.ds["usa_wind"].values[index, it]
            pc = self.ds["usa_pres"].values[index, it]
            RMW = self.ds["usa_rmw"].values[index, it]
            R35_NE = self.ds["usa_r34"].values[index, it, 0]
            R35_SE = self.ds["usa_r34"].values[index, it, 1]
            R35_SW = self.ds["usa_r34"].values[index, it, 2]
            R35_NW = self.ds["usa_r34"].values[index, it, 3]
            R50_NE = self.ds["usa_r50"].values[index, it, 0]
            R50_SE = self.ds["usa_r50"].values[index, it, 1]
            R50_SW = self.ds["usa_r50"].values[index, it, 2]
            R50_NW = self.ds["usa_r50"].values[index, it, 3]
            R65_NE = self.ds["usa_r64"].values[index, it, 0]
            R65_SE = self.ds["usa_r64"].values[index, it, 1]
            R65_SW = self.ds["usa_r64"].values[index, it, 2]
            R65_NW = self.ds["usa_r64"].values[index, it, 3]
            # R100_NE = self.ds["usa_r100"].values[index, it, 0]
            # R100_SE = self.ds["usa_r100"].values[index, it, 1]
            # R100_SW = self.ds["usa_r100"].values[index, it, 2]
            # R100_NW = self.ds["usa_r100"].values[index, it, 3]
            vmax = np.nan_to_num(vmax, copy=True, nan=-999.0)
            pc = np.nan_to_num(pc, copy=True, nan=-999.0)
            RMW = np.nan_to_num(RMW, copy=True, nan=-999.0)
            R35_NE = np.nan_to_num(R35_NE, copy=True, nan=-999.0)
            R35_SE = np.nan_to_num(R35_SE, copy=True, nan=-999.0)
            R35_SW = np.nan_to_num(R35_SW, copy=True, nan=-999.0)
            R35_NW = np.nan_to_num(R35_NW, copy=True, nan=-999.0)
            R50_NE = np.nan_to_num(R50_NE, copy=True, nan=-999.0)
            R50_SE = np.nan_to_num(R50_SE, copy=True, nan=-999.0)
            R50_SW = np.nan_to_num(R50_SW, copy=True, nan=-999.0)
            R50_NW = np.nan_to_num(R50_NW, copy=True, nan=-999.0)
            R65_NE = np.nan_to_num(R65_NE, copy=True, nan=-999.0)
            R65_SE = np.nan_to_num(R65_SE, copy=True, nan=-999.0)
            R65_SW = np.nan_to_num(R65_SW, copy=True, nan=-999.0)
            R65_NW = np.nan_to_num(R65_NW, copy=True, nan=-999.0)
            R100_NE = -999.0
            R100_SE = -999.0
            R100_SW = -999.0
            R100_NW = -999.0

            # Create a geopandas dataframe
            gdf_point = gpd.GeoDataFrame(
                {
                    "datetime": [tc_time_string],
                    "geometry": [point],
                    "vmax": [vmax],
                    "pc": [pc],
                    "rmw": [RMW],
                    "r35_ne": [R35_NE],
                    "r35_se": [R35_SE],
                    "r35_sw": [R35_SW],
                    "r35_nw": [R35_NW],
                    "r50_ne": [R50_NE],
                    "r50_se": [R50_SE],
                    "r50_sw": [R50_SW],
                    "r50_nw": [R50_NW],
                    "r65_ne": [R65_NE],
                    "r65_se": [R65_SE],
                    "r65_sw": [R65_SW],
                    "r65_nw": [R65_NW],
                    "r100_ne": [R100_NE],
                    "r100_se": [R100_SE],
                    "r100_sw": [R100_SW],
                    "r100_nw": [R100_NW],
                }
            )

            # Append self
            gdf = pd.concat([gdf, gdf_point])

        # Replace -999.0 with NaN
        gdf = gdf.replace(-999.0, np.nan)
        gdf = gdf.reset_index(drop=True)
        gdf = gdf.set_crs(crs=4326, inplace=True)

        tc.track.gdf = gdf

        return tc

    def list_names(self, index=None):
        if index is not None:
            return [self.storm_name[i] for i in index]
        else:
            return self.storm_name

    def to_gdf(self, index=None):
        """
        Returns a geopandas dataframe of the tracks.

        Parameters
        ----------
        index : list
            List of indices of the tracks to return. If None, all tracks are returned.

        Returns
        -------
        gdf : geopandas.GeoDataFrame
            Geopandas dataframe of the tracks.
        """

        # Initialize
        geom = []
        iok = []
        if index is None:
            index = range(self.nstorms)
        description = []

        # Loop over tracks
        for ind in index:
            lon = self.lon[ind, :]
            lat = self.lat[ind, :]
            lon = lon[np.isfinite(lon)]
            lat = lat[np.isfinite(lat)]
            if len(lon) > 1 and len(lat) > 1:
                geom.append(
                    shapely.geometry.LineString(np.transpose(np.stack((lon, lat))))
                )
                iok.append(ind)
                description.append(self.storm_name[ind] + " (" + str(self.year[ind]) + ")")

        # Create geopandas dataframe
        gdf = gpd.GeoDataFrame(crs=4326, geometry=geom)
        names = [self.storm_name[i] for i in iok]
        years = self.year[iok]
        gdf["name"] = names
        gdf["year"] = years
        gdf["description"] = description
        gdf["dataset_index"] = iok
        return gdf

    def filter(
        self,
        name=None,
        distance=None,
        lon=None,
        lat=None,
        basin=None,
        year=None,
        year_min=None,
        year_max=None,
        vmax_min=None,
        vmax_max=None,
    ):
        """
        Filter the dataset and return the indices of the filtered tracks.

        Parameters
        ----------
        name : str
            Name of the track to filter by.
        distance : float
            Distance (km) of the track to the point (lon, lat).
        lon : list
            Longitude range of the tracks.
        lat : list
            Latitude range of the tracks.
        basin : str
            Basin of the tracks.
        year : int
            Year of the tracks.
        year_min : int
            Minimum year of the tracks.
        year_max : int
            Maximum year of the tracks.

        Returns
        -------
        index : list
            List of indices of the filtered tracks.
        """

        # Initialize year_min and year_max
        if year:
            year_min = year
            year_max = year
        else:
            if not year_min:
                year_min = 0
            if not year_max:
                year_max = 9999

        if not vmax_min:
            vmax_min = 0.0
        if not vmax_max:
            vmax_max = 9999.0

        # Filter by basin
        if basin:
            ibasin = np.array(
                [
                    i
                    for i in range(len(self.basins))
                    if self.basin[i].lower() == basin.lower()
                ]
            )
        else:
            ibasin = np.arange(0, self.nstorms)

        # Filter by year
        if year_min and year_max:
            iyear = np.where((self.year >= year_min) & (self.year <= year_max))[0]
        else:
            iyear = np.arange(0, self.nstorms)

        # # Filter by vmax
        # if vmax_min and vmax_max:
        #     ivmax = np.where((self.year >= year_min) & (self.year <= year_max))[0]
        # else:
        #     iyear = np.arange(0, self.nstorms)

        # Filter by name
        if name:
            iname = np.array(
                [
                    i
                    for i in range(len(self.storm_name))
                    if self.storm_name[i].lower() == name.lower()
                ]
            )
        else:
            iname = np.arange(0, self.nstorms)

        # Filter by bounding box
        if isinstance(lon, list) and isinstance(lat, list):
            inear = np.where(
                (self.lon >= lon[0])
                & (self.lon <= lon[1])
                & (self.lat >= lat[0])
                & (self.lat <= lat[1])
            )
            ibbox = np.unique(inear[0])
        else:
            ibbox = np.arange(0, self.nstorms)

        # Filter by distance
        if distance:
            # Compute distance of all tracks to point
            d = CycloneTrackDataset.compute_distance(lon, lat, self.lon, self.lat)
            dmin = np.nanmin(d, axis=1)
            idist = np.where(dmin < distance)[0]
        else:
            idist = np.arange(0, self.nstorms)

        # Intersect all filters
        index = reduce(np.intersect1d, (ibasin, iyear, iname, idist, ibbox))

        return index

    @staticmethod
    def compute_distance(lon1, lat1, lon2, lat2):
        """
        Compute the distance between two points on a sphere.

        Parameters
        ----------
        lon1 : numpy.ndarray
            Longitude of the first point.
        lat1 : numpy.ndarray
            Latitude of the first point.
        lon2 : numpy.ndarray
            Longitude of the second point.
        lat2 : numpy.ndarray
            Latitude of the second point.

        Returns
        -------
        d : numpy.ndarray
            Distance between the points.
        """

        # Convert to radians
        R = 6373.0
        lon1 = lon1 * np.pi / 180
        lat1 = lat1 * np.pi / 180
        lon2 = lon2 * np.pi / 180
        lat2 = lat2 * np.pi / 180
        dlon = lon2 - lon1

        # Wrap around
        dlon[np.where(dlon < -np.pi)] = dlon[np.where(dlon < -np.pi)] + 2 * np.pi
        dlon[np.where(dlon > np.pi)] = dlon[np.where(dlon > np.pi)] - 2 * np.pi
        dlat = lat2 - lat1

        # Compute distance
        a = (np.sin(dlat / 2)) ** 2 + np.cos(lat1) * np.cos(lat2) * (
            np.sin(dlon / 2)
        ) ** 2
        c = 2 * np.arctan2(np.sqrt(a), np.sqrt(1.0 - a))
        return R * c



# Load modules
import os
from datetime import datetime

from cht_cyclones.tropical_cyclone import TropicalCyclone, TropicalCycloneEnsemble

# Define folder and track
tc                  = TropicalCyclone(name="Idai")
current_directory   = os.path.dirname(os.path.abspath(__file__))

# Read a track
file_path           = os.path.join(current_directory, 'best_track_idai.cyc')
tc.from_ddb_cyc(file_path)
tc.account_for_forward_speed()
tc.estimate_missing_values()
tc.include_rainfall = True

# Forecasting
# Uses DeMaria et al. 2009 formulations
tc2                     = TropicalCycloneEnsemble(name="Idai_forecastng", TropicalCyclone=tc)
tc2.dt                  = 3                             # 3 hourly
tc2.include_best_track  = 0                             # not included best-track
tc2.tstart              = datetime(2019,3,9,0,0,0)      # March 5, 2019 at midnight UTC
tc2.tend                = datetime(2019,3,16,6,0,0)     # last timestamp in cyc
nm_to_m                 = float(1.852) * 1000           # nm to m
tc2.mean_abs_cte24      = 19.0397*nm_to_m               # mean absolute error in cross-track error (CTE) in meter
tc2.sc_cte              = 1.3253                        # auto-regression CTE; typically >1
tc2.compute_ensemble(10)

# Write out
tc2.to_shapefile(current_directory)
tc2.to_spiderweb(current_directory)
tc2.make_figures(current_directory)


import os
from datetime import datetime, timedelta

import numpy as np
import pandas as pd
from geojson import Feature, FeatureCollection
from geopandas import GeoDataFrame
from pyproj import CRS, Geod
from scipy.interpolate import CubicSpline, interp1d
from shapely.geometry import LineString, MultiLineString, Point, mapping

from .utils import gdf_to_geojson_js, gdf_to_pli
from .wind_profiles import wind_radii_nederhoff, wpr_holland2008

knots_to_ms = float(0.51444)
nm_to_km = float(1.852)
nm_to_m = float(1.852) * 1000
dateformat_module = "%Y%m%d %H%M%S"

geodesic = Geod(ellps="WGS84")


class TropicalCycloneTrack:
    def __init__(self):
        self.gdf = GeoDataFrame()
        self.crs = CRS.from_epsg(4326)
        self.unit_intensity = "knots"
        self.unit_radii = "nm"

    def read(self, filename, tau=0):
        """Read a tropical cyclone track from a file."""

        # Try to read cyc file new format
        try:
            gdf = read_cyc(filename)
            self.gdf = gdf
            self.apply_tau(tau)
            return None
        except:
            pass

        # Try to read trk file (COAMPS-TC)
        try:
            gdf = read_trk(filename)
            self.gdf = gdf
            # Get rid of NaN values in vmax
            self.fix_nan_vmax()
            self.apply_tau(tau)
            return None
        except:
            pass

        # Try to read ddb cyc file
        try:
            gdf, config, name = read_ddb_cyc(filename)
            self.gdf = gdf
            self.apply_tau(tau)
            # Do not do anything with name
            return config
        except:
            pass

        raise Exception("Error: could not read file " + filename)

    def apply_tau(self, tau):
        """Apply tau to the track, i.e. start later if timec in track is smaller than tau"""
        if tau > 0:
            tstart = datetime.strptime(self.gdf.datetime[0], dateformat_module) + timedelta(hours=tau)
            self.shorten(tstart=tstart)

    def fix_nan_vmax(self):
        # Sometimes COAMPS-TC data has 0.0 for vmax, which is replaced by NaN
        # This is causes problems later on. Try to fix it by doing an interpolation where necessary. If it happens at the beginning or end, use nearest neighbor.
        okay = True
        for it in range(len(self.gdf)):
            if np.isnan(self.gdf.vmax[it]):
                okay = False
                break

        if not okay:
            # Make pandas series
            s = pd.Series(self.gdf.vmax)
            # Interpolate
            s = s.interpolate(method="linear")
            # Put values back in gdf
            for it in range(len(self.gdf)):
                if np.isnan(self.gdf.vmax[it]):
                    print("Warning! Replacing vmax = NaN in track with interpolated value at it = " + str(it))
                    self.gdf.loc[it, "vmax"] = s.values[it]

    def write(self,
              filename,
              format="cyc",
              config=None,
              name=None,
              include_header=True):

        # Make a copy of the track and replace NaN with -999.0
        gdf = self.gdf.fillna(-999.0)

        if format == "ddb_cyc":
            if config is None:
                print("Error: no configuration file provided")
                return
            write_ddb_cyc(filename, gdf, config, name, self.unit_intensity, self.unit_radii)
        elif format == "cyc":
            write_cyc(filename, gdf, include_header=include_header)
        else:
            raise Exception("Error: track format " + format + " not supported")

    # 3A. convert_units_imperial_metric
    def convert_units_imperial_to_metric(self, config):
        # Convert wind speeds
        # from  knots   - typically 1-minute averaged
        # to    m/s     - we account for conversion here        
        if (self.unit_intensity == "knots") and (self.unit_radii == "nm"):

            # Convert wind radii
            for it in range(len(self.gdf)):
                if self.gdf.vmax[it] > 0.0:
                    self.gdf.loc[it, "vmax"] = self.gdf.vmax[it] * knots_to_ms * config["wind_conversion_factor"]
                else:
                    self.gdf.loc[it, "vmax"] = np.nan    
                if self.gdf.rmw[it] > 0.0:
                    self.gdf.loc[it, "rmw"] = self.gdf.rmw[it] * nm_to_km
                else:
                    self.gdf.loc[it, "rmw"] = np.nan    
                # R35
                if self.gdf.r35_ne[it] > 0.0:
                    self.gdf.loc[it, "r35_ne"] = self.gdf.r35_ne[it] * nm_to_km
                else:
                    self.gdf.loc[it, "r35_ne"] = np.nan

                if self.gdf.r35_se[it] > 0.0:
                    self.gdf.loc[it, "r35_se"] = self.gdf.r35_se[it] * nm_to_km
                else:
                    self.gdf.loc[it, "r35_se"] = np.nan

                if self.gdf.r35_sw[it] > 0.0:
                    self.gdf.loc[it, "r35_sw"] = self.gdf.r35_sw[it] * nm_to_km
                else:
                    self.gdf.loc[it, "r35_sw"] = np.nan

                if self.gdf.r35_nw[it] > 0.0:
                    self.gdf.loc[it, "r35_nw"] = self.gdf.r35_nw[it] * nm_to_km
                else:
                    self.gdf.loc[it, "r35_nw"] = np.nan

                # r50
                if self.gdf.r50_ne[it] > 0.0:
                    self.gdf.loc[it, "r50_ne"] = self.gdf.r50_ne[it] * nm_to_km
                else:
                    self.gdf.loc[it, "r50_ne"] = np.nan

                if self.gdf.r50_se[it] > 0.0:
                    self.gdf.loc[it, "r50_se"] = self.gdf.r50_se[it] * nm_to_km
                else:
                    self.gdf.loc[it, "r50_se"] = np.nan

                if self.gdf.r50_sw[it] > 0.0:
                    self.gdf.loc[it, "r50_sw"] = self.gdf.r50_sw[it] * nm_to_km
                else:
                    self.gdf.loc[it, "r50_sw"] = np.nan

                if self.gdf.r50_nw[it] > 0.0:
                    self.gdf.loc[it, "r50_nw"] = self.gdf.r50_nw[it] * nm_to_km
                else:
                    self.gdf.loc[it, "r50_nw"] = np.nan

                # r65
                if self.gdf.r65_ne[it] > 0.0:
                    self.gdf.loc[it, "r65_ne"] = self.gdf.r65_ne[it] * nm_to_km
                else:
                    self.gdf.loc[it, "r65_ne"]= np.nan

                if self.gdf.r65_se[it] > 0.0:
                    self.gdf.loc[it, "r65_se"] = self.gdf.r65_se[it] * nm_to_km
                else:
                    self.gdf.loc[it, "r65_se"] = np.nan

                if self.gdf.r65_sw[it] > 0.0:
                    self.gdf.loc[it, "r65_sw"] = self.gdf.r65_sw[it] * nm_to_km
                else:
                    self.gdf.loc[it, "r65_sw"] = np.nan

                if self.gdf.r65_nw[it] > 0.0:
                    self.gdf.loc[it, "r65_nw"] = self.gdf.r65_nw[it] * nm_to_km
                else:
                    self.gdf.loc[it, "r65_nw"] = np.nan

                # r100
                if self.gdf.r100_ne[it] > 0.0:
                    self.gdf.loc[it, "r100_ne"] = self.gdf.r100_ne[it] * nm_to_km
                else:
                    self.gdf.loc[it, "r100_ne"] = np.nan

                if self.gdf.r100_se[it] > 0.0:
                    self.gdf.loc[it, "r100_se"] = self.gdf.r100_se[it] * nm_to_km
                else:
                    self.gdf.loc[it, "r100_se"] = np.nan

                if self.gdf.r100_sw[it] > 0.0:
                    self.gdf.loc[it, "r100_sw"] = self.gdf.r100_sw[it] * nm_to_km
                else:
                    self.gdf.loc[it, "r100_sw"] = np.nan

                if self.gdf.r100_nw[it] > 0.0:
                    self.gdf.loc[it, "r100_nw"] = self.gdf.r100_nw[it] * nm_to_km
                else:
                    self.gdf.loc[it, "r100_nw"] = np.nan

            # Done, so set variable
            self.unit_intensity = "ms"
            self.unit_radii = "km"

    def compute_forward_speed(self):
        """Computes forward speed and dpcdt and store these in track geodataframe"""

        # Assign variables to geopandas dataframe
        zero_list   = [0.0 for _ in range(len(self.gdf))]
        self.gdf = self.gdf.assign(vtx=zero_list)
        self.gdf = self.gdf.assign(vty=zero_list)
        self.gdf = self.gdf.assign(dpcdt=zero_list)

        # Go over time steps
        for it in range(len(self.gdf)):

            # Get basics
            datetime_it = datetime.strptime(self.gdf.datetime[it], dateformat_module)
            coords_it = self.gdf.geometry[it]
            geofacx = 1
            geofacy = 1

            # Determine geo factors (currently only geo support)
            if self.gdf.crs.name == "WGS 84":
                geofacy = 110540
                geofacx = 111320 * np.cos(coords_it.y * np.pi / 180)

            if it == 0:

                # Forward
                datetime_forward = datetime.strptime(
                    self.gdf.datetime[it + 1], dateformat_module
                )
                coords_forward = self.gdf.geometry[it + 1]
                dt = datetime_forward - datetime_it
                dt = dt.total_seconds()
                dx = (coords_forward.x - coords_it.x) * geofacx
                dy = (coords_forward.y - coords_it.y) * geofacy
                dpc = self.gdf.pc[it + 1] - self.gdf.pc[it]

            elif it == len(self.gdf) - 1:

                # Backward
                datetime_backward = datetime.strptime(
                    self.gdf.datetime[it - 1], dateformat_module
                )
                coords_backward = self.gdf.geometry[it - 1]
                dt = datetime_it - datetime_backward
                dt = dt.total_seconds()
                dx = (coords_it.x - coords_backward.x) * geofacx
                dy = (coords_it.y - coords_backward.y) * geofacy
                dpc = self.gdf.pc[it] - self.gdf.pc[it - 1]

            else:

                # Forward
                datetime_forward = datetime.strptime(
                    self.gdf.datetime[it + 1], dateformat_module
                )
                coords_forward = self.gdf.geometry[it + 1]
                dt1 = datetime_forward - datetime_it
                dt1 = dt1.total_seconds()
                dx1 = (coords_forward.x - coords_it.x) * geofacx
                dy1 = (coords_forward.y - coords_it.y) * geofacy
                dpc1 = self.gdf.pc[it + 1] - self.gdf.pc[it]

                # Backward
                datetime_backward = datetime.strptime(
                    self.gdf.datetime[it - 1], dateformat_module
                )
                coords_backward = self.gdf.geometry[it - 1]
                dt2 = datetime_it - datetime_backward
                dt2 = dt2.total_seconds()
                dx2 = (coords_it.x - coords_backward.x) * geofacx
                dy2 = (coords_it.y - coords_backward.y) * geofacy
                dpc2 = self.gdf.pc[it] - self.gdf.pc[it - 1]

                # Combined yields central differences
                dx = np.mean([dx1, dx2])
                dy = np.mean([dy1, dy2])
                dt = np.mean([dt1, dt2])
                dpc = np.mean([dpc1, dpc2])

            # Check dt so we do not divide by zero (always assume a minimum of 10 min)
            if dt < 600.0:
                dt = 600.0

            # Compute variables
            ux = dx / dt  # speed in meter per seconds
            uy = dy / dt
            dpcdt = dpc / (dt / 3600)  # pressure change per hour

            # Check to limit dpc which happens when pc is not know
            if dpcdt > 100.0 or dpcdt < -100.0 or np.isnan(dpcdt):
                dpcdt = 0.0

            # Save as part of the track
            self.gdf.vtx[it] = ux  # forward speed in x
            self.gdf.vty[it] = uy  # forward speed in y
            self.gdf.dpcdt[it] = dpcdt  # pressure difference in time

    # Support functions for creating spiderweb
    # 1. estimate_missing_values => still assuming imperial system
    def estimate_missing_values(self, config):
        # Go over the track and determine missing values

        for it in range(len(self.gdf)):
            # Get coordinates
            coords_it = self.gdf.geometry[it]

            # determine wind speed
            if np.isnan(self.gdf.vmax[it]):
                if config["wind_pressure_relation"] == "holland2008":
                    # estimate this: vmax is in m/s
                    vmax = wpr_holland2008(
                        pc=self.gdf.pc[it],
                        pn=config["background_pressure"],
                        phi=coords_it.y,
                        dpcdt=self.gdf.dpcdt[it],
                        vt=np.sqrt(self.gdf.vtx[it] ** 2 + self.gdf.vty[it] ** 2),
                        rhoa=config["rho_air"],
                    )

                    # # place this
                    # if self.unit_intensity == "knots":
                    #     self.gdf.vmax[it] = vmax / knots_to_ms
                    # else:
                    self.gdf.vmax[it] = vmax

            # determine pressure
            if np.isnan(self.gdf.pc[it]):
                if config["wind_pressure_relation"] == "holland2008":
                    # estimate this
                    if self.unit_intensity == "knots":
                        pc = wpr_holland2008(
                            vmax=self.gdf.vmax[it] * knots_to_ms,
                            pn=self.background_pressure,
                            phi=coords_it.y,
                            dpcdt=self.gdf.dpcdt[it],
                            vt=np.sqrt(
                                self.gdf.vtx[it] ** 2 + self.gdf.vty[it] ** 2
                            ),
                            rhoa=self.rho_air,
                        )
                    else:
                        pc = wpr_holland2008(
                            vmax=self.gdf.vmax[it],
                            pn=config["background_pressure"],
                            phi=coords_it.y,
                            dpcdt=self.gdf.dpcdt[it],
                            vt=np.sqrt(
                                self.gdf.vtx[it] ** 2 + self.gdf.vty[it] ** 2
                            ),
                            rhoa=config["rho_air"],
                        )

                    # place this
                    self.gdf.pc[it] = pc

            # radius of maximum winds (RMW)
            if np.isnan(self.gdf.rmw[it]):
                # Nederhoff et al. 2019
                if config["rmw_relation"] == "nederhoff2019":
                    # Estimate: relationship assumes m/s
                    if self.unit_intensity == "knots":
                        [rmax, dr35] = wind_radii_nederhoff(
                            self.gdf.vmax[it] * knots_to_ms, coords_it.y, 7, 0
                        )
                    else:
                        [rmax, dr35] = wind_radii_nederhoff(
                            self.gdf.vmax[it], coords_it.y, 7, 0
                        )

                    # Place value: output is in km
                    if self.unit_radii == "nm":
                        self.gdf.rmw[it] = rmax["mode"] / nm_to_km
                    else:
                        self.gdf.rmw[it] = rmax["mode"]

                # Gross et al. 2004
                elif config["rmw_relation"] == "gross2004":
                    # Estimate: relationship assume knots
                    if self.unit_radii == "knots":
                        rmax = (
                            35.37
                            - 0.11100 * self.gdf.vmax[it]
                            + 0.5700 * (abs(coords_it.y) - 25)
                        )
                    else:
                        rmax = (
                            35.37
                            - 0.11100 * self.gdf.vmax[it] / knots_to_ms
                            + 0.5700 * (abs(coords_it.y) - 25)
                        )

                    # Place value: output is in nm
                    if self.unit_radii == "nm":
                        self.gdf.rmw[it] = rmax
                    else:
                        self.gdf.rmw[it] = rmax * nm_to_km

                # Simple constant value of 25 nm
                elif config["rmw_relation"] == "constant_25nm":
                    # Place 25 nm
                    if self.unit_radii == "nm":
                        self.gdf.rmw[it] = float(25)
                    else:
                        self.gdf.rmw[it] = float(25) * nm_to_km

            # radius of gale force winds (R35)
            if config["wind_profile"] == "holland2010":
                if self.gdf.vmax[it] >= 35:
                    if (
                        (self.gdf.r35_ne[it] == -999)
                        and (self.gdf.r35_se[it] == -999)
                        and (self.gdf.r35_sw[it] == -999)
                        and (self.gdf.r35_nw[it] == -999)
                    ):
                        # Estimate values
                        if self.unit_radii == "knots":
                            [rmax, dr35] = wind_radii_nederhoff(
                                self.gdf.vmax[it] * knots_to_ms, coords_it.y, 7, 0
                            )
                        else:
                            [rmax, dr35] = wind_radii_nederhoff(
                                self.gdf.vmax[it], coords_it.y, 7, 0
                            )

                        if self.unit_radii == "nm":
                            self.gdf.r35_NE[it] = (
                                dr35["mode"] / nm_to_km + self.gdf.rmw[it]
                            )
                            self.gdf.r35_se[it] = (
                                dr35["mode"] / nm_to_km + self.gdf.rmw[it]
                            )
                            self.gdf.r35_sw[it] = (
                                dr35["mode"] / nm_to_km + self.gdf.rmw[it]
                            )
                            self.gdf.r35_nw[it] = (
                                dr35["mode"] / nm_to_km + self.gdf.rmw[it]
                            )
                        else:
                            self.gdf.r35_NE[it] = dr35["mode"] + self.gdf.rmw[it]
                            self.gdf.r35_se[it] = dr35["mode"] + self.gdf.rmw[it]
                            self.gdf.r35_sw[it] = dr35["mode"] + self.gdf.rmw[it]
                            self.gdf.r35_nw[it] = dr35["mode"] + self.gdf.rmw[it]

    # # 2A. cut_off_low_wind_speeds (should not do this anymore)
    # def cut_off_low_wind_speeds(self):
    #     # Only apply this when the cut_off wind is defined
    #     if self.low_wind_speeds_cut_off > 0.0:
    #         # Find first
    #         ifirst = []
    #         for it in range(len(self.track)):
    #             if self.track.vmax[it] >= self.low_wind_speeds_cut_off and not ifirst:
    #                 ifirst = it
    #                 break
    #         if ifirst > 0:
    #             self.track = self.track.drop(list(range(0, ifirst)))
    #             self.track = self.track.reset_index(drop=True)

    #         # Find last
    #         ilast = []
    #         for it in range(len(self.track) - 1, 0 - 1, -1):
    #             if self.track.vmax[it] >= self.low_wind_speeds_cut_off and not ilast:
    #                 ilast = it
    #                 break
    #         if ilast:
    #             self.track = self.track.drop(list(range(ilast + 1, len(self.track))))
    #             self.track = self.track.reset_index(drop=True)

    #     else:
    #         if self.debug == 1:
    #             print("No cut_off_low_wind_speeds since wind speed is zero or lower")

    def extend(self, hours_before=None, hours_after=None):
        """Extend the track with a certain number of hours before and after the track."""
        if hours_before is not None:
            t0 = datetime.strptime(self.gdf.datetime[0], dateformat_module)
            t = t0 - timedelta(hours=hours_before)
            self.add_point(t)
        if hours_after is not None:
            t1 = datetime.strptime(self.gdf.datetime[len(self.gdf) - 1], dateformat_module)
            t = t1 + timedelta(hours=hours_after)
            self.add_point(t)    

    def shorten(self, tstart=None, tend=None):
        """Shorten the track to a certain time period."""
        times_to_remove = []
        if tstart is not None:
            for it, row in self.gdf.iterrows():
                t = datetime.strptime(row.datetime, dateformat_module)
                if t < tstart:
                    times_to_remove.append(t)
                    # self.remove_point(time=t)
        if tend is not None:
            for it, row in self.gdf.iterrows():
                t = datetime.strptime(row.datetime, dateformat_module)
                if t > tend:
                    times_to_remove.append(t)
                    # self.remove_point(time=t)            
        if len(times_to_remove) > 0:
            for t in times_to_remove:
                self.remove_point(time=t)

    def resample(self, dt, method="spline"):

        """Resample the track to a new time step. Returns new gdf. By default uses a spline interpolation for the track points."""
        
        # Get the first and last time
        t0 = datetime.strptime(self.gdf.datetime[0], dateformat_module)
        t1 = datetime.strptime(self.gdf.datetime[len(self.gdf) - 1], dateformat_module)
        dt = timedelta(hours=dt)

        # Determine number of time steps
        nt = int((t1 - t0) / dt) + 1

        # Create a new track gdf
        gdf = GeoDataFrame()

        # Copy first point of existing track
        for it in range(nt):
            gdf = pd.concat([gdf, self.gdf.iloc[[0]]], ignore_index=True)
            gdf.loc[it, "datetime"] = (t0 + it * dt).strftime(dateformat_module)

        self.gdf = interpolate_track(self.gdf, gdf, method=method)

        return gdf

    def add_point(self, time: datetime,
                  lon: float=-9999.0,
                  lat: float=-9999.0,
                  method: str="spline"):
        """Insert a point at a certain time and location (optional)"""
            
        # Find index where to insert
        # Check first if time is before the first time
        t0 = datetime.strptime(self.gdf.datetime[0], dateformat_module)
        if time < t0:
            index = 0
        else:
            for it, row in self.gdf.iterrows():
                if time == datetime.strptime(row.datetime, dateformat_module):
                    raise Exception("Error: time already exists in track")
                if time > datetime.strptime(row.datetime, dateformat_module):
                    index = it + 1

        if index == 0 or index == len(self.gdf):
            # Need to compute the forward speed to estimate track position
            self.compute_forward_speed()

        gdf_point, index = interpolate_to_point(self.gdf, time, method=method)

        # Check if lon and lat are provided
        if lon == -9999.0 or lat == -9999.0:
            # Not provided so use interpolated values
            pass
        else:
            # Use provided coordinates
            point = Point(lon, lat)
            gdf_point["geometry"] = [point]

        gdf_point.replace(-999.0, np.nan, inplace=True)
        crs = self.gdf.crs

        if index == 0:
            self.gdf = pd.concat([gdf_point, self.gdf], ignore_index=True)
        elif index == len(self.gdf):
            self.gdf = pd.concat([self.gdf, gdf_point], ignore_index=True)
        else:
            self.gdf = pd.concat([self.gdf.iloc[:index], gdf_point, self.gdf.iloc[index:]], ignore_index=True)            

        self.gdf = self.gdf.set_crs(crs)    
        self.gdf = self.gdf.reset_index(drop=True)

    def remove_point(self, index=None, time=None):
        """Remove point from the track by index or time"""

        if index is not None:
            self.gdf = self.gdf.drop(index)
            self.gdf = self.gdf.reset_index(drop=True)

        if time is not None:
            # Loop through the track and find the index
            for index, row in self.gdf.iterrows():
                t = datetime.strptime(row.datetime, dateformat_module)
                if t == time:
                    self.gdf = self.gdf.drop(index)
                    self.gdf = self.gdf.reset_index(drop=True)
                    break

    def add(self, track1):
        """Add another track to the current track"""

        # If self is empty, just copy track1
        if len(self.gdf) == 0:
            self.gdf = track1.gdf
            return

        # Get first time of track1
        t0 = datetime.strptime(track1.gdf.datetime[0], dateformat_module)

        # Remove all points in self that are at or after t0
        for it, row in self.gdf.iterrows():
            t = datetime.strptime(row.datetime, dateformat_module)
            if t >= t0:
                self.remove_point(time=t)

        # Concatenate the two tracks
        self.gdf = pd.concat([self.gdf, track1.gdf], ignore_index=True).reset_index(drop=True)


    def to_gdf(self, filename=None):
        """Make track GeoDataFrame and optionally write to file"""

        categories = {33.0: "TD", 64.0: "TS", 83.0: "1", 96.0: "2", 113.0: "3", 137.0: "4"}

        features = []
        points = []

        # First the track
        for ip in range(np.size(self.gdf.geometry.x)):
            points.append([self.gdf.geometry.x[ip], self.gdf.geometry.y[ip]])
        features.append(Feature(geometry=LineString(coordinates=points), properties={"name": "No name"}))

        # Then the points
        for ip in range(np.size(self.gdf.geometry.x)):
            point = Point((self.gdf.geometry.x[ip], self.gdf.geometry.y[ip]))
            tmptime = datetime.strptime(self.gdf.datetime[ip], "%Y%m%d %H%M%S")
            vmax = self.gdf.vmax[ip]
            # if vmax is NaN, set vmax to 1.0
            if np.isnan(vmax):
                vmax = 1.0
            for threshold, cat in categories.items():
                if vmax < threshold:
                    break
            else:
                cat = "5"
            features.append(
                Feature(
                    geometry=point,
                    properties={
                        "time": tmptime.strftime("%Y/%m/%d %H:%M") + " UTC",
                        "lon": self.gdf.geometry.x[ip],
                        "lat": self.gdf.geometry.y[ip],
                        "vmax": self.gdf.vmax[ip],
                        "pc": self.gdf.pc[ip],
                        "category": cat,
                    },
                )
            )

        gdf = GeoDataFrame().from_features(FeatureCollection(features))

        # Write to file 
        if filename is not None:
            # Get extension of filename
            ext = os.path.splitext(filename)[-1]
            # Get path of filename
            folder_path = os.path.dirname(filename)
            # Make path (if needed)
            if folder_path != "":
                os.makedirs(folder_path, exist_ok=True)
            if ext == ".shp":
                gdf.to_file(filename)
            elif ext == ".geojson":
                gdf.to_file(filename, driver="GeoJSON")
            elif ext == ".js":
                gdf_to_geojson_js(gdf, filename, varname="track_data")   
            elif ext == ".pli":
                gdf_to_pli(gdf, filename)   

        return gdf    


def read_cyc(filename):
    # cyc format

    gdf = GeoDataFrame()

    # Read all the lines first
    with open(filename, "r") as f:
        lines = f.readlines()

    for line in lines:
        if line[0] == "#":
            continue
        # Get values
        line = line.split()
        date_format = "%Y%m%d %H%M%S"
        date_string = line[0] + " " + line[1]
        tc_time = datetime.strptime(date_string, date_format)
        tc_time_string = tc_time.strftime(date_format)
        y = float(line[2])
        x = float(line[3])
        vmax = float(line[4])
        pc = float(line[5])
        RMW = float(line[6])
        R35_NE = float(line[7])
        R35_SE = float(line[8])
        R35_SW = float(line[9])
        R35_NW = float(line[10])
        R50_NE = float(line[11])
        R50_SE = float(line[12])
        R50_SW = float(line[13])
        R50_NW = float(line[14])
        R65_NE = float(line[15])
        R65_SE = float(line[16])
        R65_SW = float(line[17])
        R65_NW = float(line[18])
        R100_NE = float(line[19])
        R100_SE = float(line[20])
        R100_SW = float(line[21])
        R100_NW = float(line[22])

        # Make GeoDataFrame
        point = Point(x, y)
        gdf_point = GeoDataFrame(
            {
                "datetime": [tc_time_string],
                "geometry": [point],
                "vmax": [vmax],
                "pc": [pc],
                "rmw": [RMW],
                "r35_ne": [R35_NE],
                "r35_se": [R35_SE],
                "r35_sw": [R35_SW],
                "r35_nw": [R35_NW],
                "r50_ne": [R50_NE],
                "r50_se": [R50_SE],
                "r50_sw": [R50_SW],
                "r50_nw": [R50_NW],
                "r65_ne": [R65_NE],
                "r65_se": [R65_SE],
                "r65_sw": [R65_SW],
                "r65_nw": [R65_NW],
                "r100_ne": [R100_NE],
                "r100_se": [R100_SE],
                "r100_sw": [R100_SW],
                "r100_nw": [R100_NW],
            }
        )

        # Append self
        gdf = pd.concat([gdf, gdf_point])

    # Replace -999.0 with NaN
    gdf = gdf.replace(-999.0, np.nan)
    gdf = gdf.reset_index(drop=True)
    gdf = gdf.set_crs(crs=CRS(4326), inplace=True)

    return gdf

def read_ddb_cyc(filename):
    # Old ddb cyc format

    gdf = GeoDataFrame()
    config = {}

    # Read all the lines first
    with open(filename, "r") as f:
        lines = f.readlines()

    # Define the name first
    for line in lines:
        if line[0:4] == "Name":
            string_value = line[5:]
            string_value = "".join(ch for ch in string_value if ch.isalnum())
            name = string_value

    # Define other variables names (if they exist)
    for i in range(len(lines)):
        line = lines[i]
        if line[0:11] == "WindProfile":
            string_value = line[23:]
            string_value = "".join(ch for ch in string_value if ch.isalnum())
            config["wind_profile"] = string_value
        if line[0:20] == "WindPressureRelation":
            string_value = line[23:]
            string_value = "".join(ch for ch in string_value if ch.isalnum())
            config["wind_pressure_relation"] = string_value
        if line[0:12] == "RMaxRelation":
            string_value = line[23:]
            string_value = "".join(ch for ch in string_value if ch.isalnum())
            config["rmw_relation"] = string_value
        if line[0:18] == "Backgroundpressure":
            string_value = line[23:]
            config["background_pressure"] = float(string_value)
        if line[0:9] == "PhiSpiral":
            string_value = line[23:]
            config["phi_spiral"] = float(string_value)
        if line[0:20] == "WindConversionFactor":
            string_value = line[23:]
            config["wind_conversion_factor"] = float(string_value)
        if line[0:15] == "SpiderwebRadius":
            string_value = line[23:]
            config["spiderweb_radius"] = float(string_value)
        if line[0:12] == "NrRadialBins":
            string_value = line[23:]
            config["nr_radial_bins"] = int(string_value)
        if line[0:17] == "NrDirectionalBins":
            string_value = line[23:]
            config["nr_directional_bins"] = int(string_value)

    # Read the track and find last comment line
    for i in range(len(lines)):
        line = lines[i]
        if line[0] == "#":
            icomment = i

    # Place coordinates in Tropical Cyclone Track
    for j in range(icomment + 1, len(lines)):
        # Get values
        line = lines[j]
        line = line.split()
        date_format = "%Y%m%d %H%M%S"
        date_string = line[0] + " " + line[1]
        tc_time = datetime.strptime(date_string, date_format)
        tc_time_string = tc_time.strftime(date_format)
        y = float(line[2])
        x = float(line[3])
        vmax = float(line[4])
        pc = float(line[5])
        RMW = float(line[6])
        R35_NE = float(line[7])
        R35_SE = float(line[8])
        R35_SW = float(line[9])
        R35_NW = float(line[10])
        R50_NE = float(line[11])
        R50_SE = float(line[12])
        R50_SW = float(line[13])
        R50_NW = float(line[14])
        R65_NE = float(line[15])
        R65_SE = float(line[16])
        R65_SW = float(line[17])
        R65_NW = float(line[18])
        R100_NE = float(line[19])
        R100_SE = float(line[20])
        R100_SW = float(line[21])
        R100_NW = float(line[22])

        # Make GeoDataFrame
        point = Point(x, y)
        gdf_point = GeoDataFrame(
            {
                "datetime": [tc_time_string],
                "geometry": [point],
                "vmax": [vmax],
                "pc": [pc],
                "rmw": [RMW],
                "r35_ne": [R35_NE],
                "r35_se": [R35_SE],
                "r35_sw": [R35_SW],
                "r35_nw": [R35_NW],
                "r50_ne": [R50_NE],
                "r50_se": [R50_SE],
                "r50_sw": [R50_SW],
                "r50_nw": [R50_NW],
                "r65_ne": [R65_NE],
                "r65_se": [R65_SE],
                "r65_sw": [R65_SW],
                "r65_nw": [R65_NW],
                "r100_ne": [R100_NE],
                "r100_se": [R100_SE],
                "r100_sw": [R100_SW],
                "r100_nw": [R100_NW],
            }
        )

        # Append self
        gdf = pd.concat([gdf, gdf_point])

    # Replace -999.0 with NaN
    gdf = gdf.replace(-999.0, np.nan)
    gdf = gdf.reset_index(drop=True)
    gdf = gdf.set_crs(crs=CRS(4326), inplace=True)

    return gdf, config, name

def read_trk(filename):
    # Read the track from a TRK file

    # Create a new empty GDF
    gdf = GeoDataFrame()

    # Initialize variables
    lasttime = np.datetime64(datetime.strptime("2000010118", "%Y%m%d%H"))

    # Open the file
    with open(filename, "r") as fid:

        # Read line by line
        for s0 in fid:
            s0 = s0.strip()
            if not s0:
                continue

            vmax = np.nan
            pc   = np.nan
            rmax = np.nan
            r35_ne = np.nan
            r35_se = np.nan
            r35_sw = np.nan
            r35_nw = np.nan
            r50_ne = np.nan
            r50_se = np.nan
            r50_sw = np.nan
            r50_nw = np.nan
            r65_ne = np.nan
            r65_se = np.nan
            r65_sw = np.nan
            r65_nw = np.nan
            r100_ne = np.nan
            r100_se = np.nan
            r100_sw = np.nan
            r100_nw = np.nan

            # Start
            s = s0.split(",")
            tc = {"name": "not_known"}
            tc["basin"] = s[0]
            tc["storm_number"] = int(s[1])

            # Read time
            tstr = str(s[2])
            tstr = tstr.replace(" ", "")
            time = datetime.strptime(tstr, "%Y%m%d%H")
            newtime = np.datetime64(time)

            # Add forecasted time
            hrs = float(s[5])
            newtime = newtime + np.timedelta64(int(hrs * 3600), "s")

            # Latitude
            if s[6][-1] == "N":
                y = 0.1 * float(s[6][:-1])
            else:
                y = -0.1 * float(s[6][:-1])
            # Longitude
            if s[7][-1] == "E":
                x = 0.1 * float(s[7][:-1])
            else:
                x = -0.1 * float(s[7][:-1])

            # vmax
            if float(s[8]) > 0:
                vmax = float(s[8])

            # Pressure and raddi
            if len(s) > 9:
                # Pressure
                if float(s[9]) > 0:
                    pc = float(s[9])

                # Radii
                r = float(s[11])
                if r in [34, 35]:
                    r35_ne = float(s[13])
                    r35_se = float(s[14])
                    r35_sw = float(s[15])
                    r35_nw = float(s[16])
                elif r == 50:
                    r50_ne = float(s[13])
                    r50_se = float(s[14])
                    r50_sw = float(s[15])
                    r50_nw = float(s[16])
                elif r in [64, 65]:
                    r65_ne = float(s[13])
                    r65_se = float(s[14])
                    r65_sw = float(s[15])
                    r65_nw = float(s[16])
                elif r == 100:
                    r100_ne = float(s[13])
                    r100_se = float(s[14])
                    r100_sw = float(s[15])
                    r100_nw = float(s[16])

                # Other things
                if len(s) > 17:
                    if s[17]:
                        pressure_last_closed_isobar = float(s[17])
                    if s[18]:
                        radius_last_closed_isobar = float(s[18])
                    if s[19]:
                        try:
                            if float(s[19]) > 0:
                                rmax = float(s[19])
                        except ValueError:
                            # print("Error: Unable to convert to float for RMW")
                            rmax = np.nan
                    if len(s) >= 28:
                        if s[27]:
                            name = s[27]

            if newtime > lasttime + np.timedelta64(1, "s"):
                # New time point found
                # create new gdf
                new_point = True
                gdf_point = GeoDataFrame(geometry=[Point(x, y)])
                # gdf_point["geometry"] = [Point(x, y)]  # Assign the new geometry
                gdf_point["datetime"] = newtime.astype("O").strftime("%Y%m%d %H%M%S")
                gdf_point["vmax"]   = np.nan
                gdf_point["pc"]     = np.nan
                gdf_point["rmw"]    = np.nan
                gdf_point["r35_ne"] = np.nan
                gdf_point["r35_se"] = np.nan
                gdf_point["r35_sw"] = np.nan
                gdf_point["r35_nw"] = np.nan
                gdf_point["r50_ne"] = np.nan
                gdf_point["r50_se"] = np.nan
                gdf_point["r50_sw"] = np.nan
                gdf_point["r50_nw"] = np.nan
                gdf_point["r65_ne"] = np.nan
                gdf_point["r65_se"] = np.nan
                gdf_point["r65_sw"] = np.nan
                gdf_point["r65_nw"] = np.nan
                gdf_point["r100_ne"] = np.nan
                gdf_point["r100_se"] = np.nan
                gdf_point["r100_sw"] = np.nan
                gdf_point["r100_nw"] = np.nan
            else:
                new_point = False
                    
            # Update all variables in gdf (when it is not the first one)                
            gdf_point["vmax"] = vmax
            gdf_point["pc"]   = pc
            gdf_point["rmw"]  = rmax
            if r35_ne > 0.0:
                gdf_point["r35_ne"] = r35_ne
            if r35_se > 0.0:
                gdf_point["r35_se"] = r35_se
            if r35_sw > 0.0:
                gdf_point["r35_sw"] = r35_sw
            if r35_nw > 0.0:
                gdf_point["r35_nw"] = r35_nw
            if r50_ne > 0.0:
                gdf_point["r50_ne"] = r50_ne
            if r50_se > 0.0:
                gdf_point["r50_se"] = r50_se
            if r50_sw > 0.0:
                gdf_point["r50_sw"] = r50_sw
            if r50_nw > 0.0:
                gdf_point["r50_nw"] = r50_nw
            if r65_ne > 0.0:
                gdf_point["r65_ne"] = r65_ne
            if r65_se > 0.0:
                gdf_point["r65_se"] = r65_se
            if r65_sw > 0.0:
                gdf_point["r65_sw"] = r65_sw
            if r65_nw > 0.0:
                gdf_point["r65_nw"] = r65_nw
            if r100_ne > 0.0:
                gdf_point["r100_ne"] = r100_ne
            if r100_se > 0.0:
                gdf_point["r100_se"] = r100_se
            if r100_sw > 0.0:
                gdf_point["r100_sw"] = r100_sw
            if r100_nw > 0.0:
                gdf_point["r100_nw"] = r100_nw

            if new_point:
                # Add new point
                gdf = pd.concat([gdf, gdf_point])
            else:
                # Replace row in dataframe with row from other dataframe
                gdf.iloc[-1] = gdf_point.iloc[0]

            lasttime = newtime

    # Done with this => rest so track looks good
    gdf = gdf.reset_index(drop=True)
    gdf = gdf.set_crs(crs=CRS(4326), inplace=True)

    return gdf

def write_cyc(filename, gdf, include_header=True):   
    """Write to cyc format"""
    with open(filename, "wt") as f:
        # Print header
        if include_header:
            f.writelines(
                "#   Date   Time      Lat      Lon     Vmax       Pc     Rmax  R35(NE)  R35(SE)  R35(SW)  R35(NW)  R50(NE)  R50(SE)  R50(SW)  R50(NW)  R65(NE)  R65(SE)  R65(SW)  R65(NW) R100(NE) R100(SE) R100(SW) R100(NE)\n"
            )

        # Print the actual track
        for i in range(len(gdf)):

            f.writelines(gdf.datetime[i].rjust(15))
            coords = gdf.geometry[i]
            f.writelines(str(round(coords.y, 2)).rjust(9))
            f.writelines(str(round(coords.x, 2)).rjust(9))

            f.writelines(str(round(gdf.vmax[i], 1)).rjust(9))
            f.writelines(str(round(gdf.pc[i], 1)).rjust(9))
            f.writelines(str(round(gdf.rmw[i], 1)).rjust(9))

            f.writelines(str(round(gdf.r35_ne[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r35_se[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r35_sw[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r35_nw[i], 1)).rjust(9))

            f.writelines(str(round(gdf.r50_ne[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r50_se[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r50_sw[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r50_nw[i], 1)).rjust(9))

            f.writelines(str(round(gdf.r65_ne[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r65_se[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r65_sw[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r65_nw[i], 1)).rjust(9))

            f.writelines(str(round(gdf.r100_ne[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r100_se[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r100_sw[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r100_nw[i], 1)).rjust(9))

            f.writelines("\n")

def write_ddb_cyc(filename, gdf, config, name, unit_intensity, unit_radii):
    """Write to 'old' Delft Dashboard cyc format"""
    with open(filename, "wt") as f:
        # Print header
        f.writelines(
            "# Tropical Cyclone Toolbox - Coastal Hazards Toolkit - "
            + datetime.now().strftime(dateformat_module)
            + "\n"
        )
        # Print rest
        f.writelines('Name                   "' + name + '"\n')
        f.writelines("WindProfile            " + config["wind_profile"] + "\n")
        f.writelines(
            "WindPressureRelation   " + config["wind_pressure_relation"] + "\n"
        )
        f.writelines("RMaxRelation           " + config["rmw_relation"] + "\n")
        f.writelines(
            "Backgroundpressure     " + str(config["background_pressure"]) + "\n"
        )
        f.writelines("PhiSpiral              " + str(config["phi_spiral"]) + "\n")
        f.writelines(
            "WindConversionFactor   " + str(config["wind_conversion_factor"]) + "\n"
        )
        f.writelines(
            "SpiderwebRadius        " + str(config["spiderweb_radius"]) + "\n"
        )
        f.writelines(
            "NrRadialBins           " + str(config["nr_radial_bins"]) + "\n"
        )
        f.writelines(
            "NrDirectionalBins      " + str(config["nr_directional_bins"]) + "\n"
        )
        f.writelines("EPSG                   WGS84\n")
        f.writelines(
            "UnitIntensity          " + str(unit_intensity) + "\n"
        )
        f.writelines("UnitWindRadii          " + str(unit_radii) + "\n")

        # Print header for the track
        f.writelines("#  \n")

        f.writelines(
            "#   Date   Time      Lat      Lon     Vmax       Pc     Rmax  R35(NE)  R35(SE)  R35(SW)  R35(NW)  R50(NE)  R50(SE)  R50(SW)  R50(NW)  R65(NE)  R65(SE)  R65(SW)  R65(NW) R100(NE) R100(SE) R100(SW) R100(NE)\n"
        )
        f.writelines("#  \n")

        # Print the actual track
        for i in range(len(gdf)):

            f.writelines(gdf.datetime[i].rjust(15))
            coords = gdf.geometry[i]
            f.writelines(str(round(coords.y, 2)).rjust(9))
            f.writelines(str(round(coords.x, 2)).rjust(9))

            f.writelines(str(round(gdf.vmax[i], 1)).rjust(9))
            f.writelines(str(round(gdf.pc[i], 1)).rjust(9))
            f.writelines(str(round(gdf.rmw[i], 1)).rjust(9))

            f.writelines(str(round(gdf.r35_ne[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r35_se[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r35_sw[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r35_nw[i], 1)).rjust(9))

            f.writelines(str(round(gdf.r50_ne[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r50_se[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r50_sw[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r50_nw[i], 1)).rjust(9))

            f.writelines(str(round(gdf.r65_ne[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r65_se[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r65_sw[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r65_nw[i], 1)).rjust(9))

            f.writelines(str(round(gdf.r100_ne[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r100_se[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r100_sw[i], 1)).rjust(9))
            f.writelines(str(round(gdf.r100_nw[i], 1)).rjust(9))

            f.writelines("\n")

def interpolate_track(gdf0, gdf1, method="linear"):
    """Interpolate track0 to track1. Only the times need to be provided in track1. The rest will be interpolated. Method is either 'linear' or 'spline'"""

    track0 = gdf_to_dict(gdf0)
    track1 = gdf_to_dict(gdf1)

    # Iterate through items in track0
    for key, value in track0.items():
        if key == "datetime" or key == "timestamp":
            continue
        elif key == "x" or key == "y":
            if method == "linear":
                f = interp1d(track0["timestamp"], value)
            else:
                f = CubicSpline(track0["timestamp"], value)
            track1[key] = f(track1["timestamp"])
        else:
            # Interpolate linearly
            f = interp1d(track0["timestamp"], value)
            track1[key] = f(track1["timestamp"])

    # And now create a new gdf
    gdf = GeoDataFrame()
    for index, time in enumerate(track1["datetime"]):
        point = Point(track1["x"][index], track1["y"][index])
        gdf_point = GeoDataFrame(geometry=[point])
        # gdf_point["geometry"] = [Point(track1["x"][index], track1["y"][index])]  # Assign the new geometry
        gdf_point["datetime"] = track1["datetime"][index].strftime(dateformat_module)
        # Loop through items in track1
        for key, value in track1.items():
            if key == "datetime" or key == "timestamp" or key == "x" or key == "y":
                continue
            else:
                gdf_point[key] = track1[key][index]
        gdf = pd.concat([gdf, gdf_point], ignore_index=True)        
    
    gdf = gdf.set_crs(gdf0.crs, inplace=True)

    return gdf

def interpolate_to_point(gdf, time: datetime, method="spline"):
    """Return interpolated track point at a certain time"""            
    # Find index where to insert
    # Check first if time is before the first time
    t0 = datetime.strptime(gdf.datetime[0], dateformat_module)
    if time < t0:
        index = 0
    else:
        for it, row in gdf.iterrows():
            if time == datetime.strptime(row.datetime, dateformat_module):
                raise Exception("Error: time already exists in track")
            if time > datetime.strptime(row.datetime, dateformat_module):
                index = it + 1

    if index == 0:
        ipos = 0
        vfac = -1.0
    elif index == len(gdf):       
        ipos = len(gdf) - 1
        vfac = 1.0

    if index == 0 or index == len(gdf):
        # Need to estimate position from forward speed and direction

        # Take starting point of original track
        gdf_point = gdf.iloc[[ipos]].copy()

        # Get time difference
        dt = np.abs((time - datetime.strptime(gdf.datetime[ipos], dateformat_module)).total_seconds())

        # Get forward speed at the start
        vtx = vfac * gdf.vtx[ipos]
        vty = vfac * gdf.vty[ipos]

        # Starting point
        lon0 = gdf.geometry[ipos].x
        lat0 = gdf.geometry[ipos].y

        # Determine azimuth and direction
        az = 90.0 - np.arctan2(vty, vtx) * 180 / np.pi
        dst = np.sqrt(vtx ** 2 + vty ** 2) * dt

        # Along track error shift
        lon1, lat1, backaz = geodesic.fwd(
            lon0,
            lat0,
            az,
            dst,
            radians=False,
        )
        point = Point(lon1, lat1)
        gdf_point["geometry"] = [point]
        gdf_point["datetime"] = [time.strftime(dateformat_module)]

    else:
        # Need to do full interpolation
        gdf_point = gdf.iloc[[0]].copy()
        gdf_point["datetime"] = [time.strftime(dateformat_module)]
        gdf_point = interpolate_track(gdf, gdf_point, method=method)

    return gdf_point, index    

def gdf_to_dict(gdf):

    track0 = {}

    track0["datetime"] = []
    track0["x"]        = []
    track0["y"]        = []
    # Loop through columns in gdf
    for column in gdf.items():
        if column[0] == "datetime":
            continue
        elif column[0] == "geometry":
            continue
        else:
            track0[column[0]] = []

    # Now loop through the times
    for index, row in gdf.iterrows():
        # Convert to datetime
        track0["datetime"].append(datetime.strptime(row.datetime, dateformat_module))
        # Get x and y
        track0["x"].append(row.geometry.x)
        track0["y"].append(row.geometry.y)
        # Loop through columns in gdf
        for column in gdf.items():
            if column[0] == "datetime":
                continue
            elif column[0] == "geometry":
                continue
            else:
                track0[column[0]].append(row[column[0]])

    # Convert to timestamp for interpolation
    track0["timestamp"] = [date.timestamp() for date in track0["datetime"]]

    return track0



# -*- coding: utf-8 -*-
"""
Tropical Cyclone Module in the Coastal Hazards Toolkit

Module supports two classes
    TropicalCyclone:             deterministic simulations
    TropicalCycloneEnsemble      probabilistic simulations using the simplified DeMaria et al. (2009) approach

    To do list - 'nice to haves'
        make reading of ddb_cyc not file size related but using actual keywords (since format are changing)
        => also remove spiderweb keywords and make it only related to the actual track (2 files?)
        add more reading formats (e.g. NHC, JTWC, etc.)
        enable coordinate conversions; now it is all WGS 84
        add more rainfall methods + allow for scaling of rainfall in ensembles
"""

# Standard Library Imports
import copy
import os
from datetime import datetime

# Third-Party Library Imports
import geopandas as gpd

# import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pyproj
import toml

from .ensemble import TropicalCycloneEnsemble
from .fit_holland_2010 import fit_wind_field_holland2010
from .spiderweb import TropicalCycloneSpiderweb
from .track import TropicalCycloneTrack
from .wind_profiles import holland2010

geodesic = pyproj.Geod(ellps="WGS84")

# Settings
dateformat_module = "%Y%m%d %H%M%S"
knots_to_ms = float(0.51444)
nm_to_km = float(1.852)
nm_to_m = float(1.852) * 1000
pd.options.mode.chained_assignment = None

# Classes of the Tropical Cyclone module
class TropicalCyclone:
    # Init
    def __init__(self,
                 name="no_name",
                 track_file=None,
                 config_file=None,
                 tau=0):
        """Initialize the Tropical Cyclone object"""

        # Name         
        self.name = name

        # Configuration  
        self.config_file = config_file
        self.set_default_config()
        if config_file is not None:
            self.read_config(config_file)

        # Track    
        self.track = TropicalCycloneTrack()
        self.track_file = track_file
        if track_file is not None:
            # Track file can be a string or a list of strings
            if isinstance(track_file, list):
                self.read_track_files_and_merge(track_file, tau=tau)
            else:
                self.read_track(track_file)

        # Spiderweb wind field
        self.spiderweb = TropicalCycloneSpiderweb()

    def set_default_config(self):
        self.config = {}
        self.config["wind_profile"] = "holland2010"
        self.config["wind_pressure_relation"] = "holland2008"
        self.config["rmw_relation"] = "nederhoff2019"
        self.config["rainfall_relationship"] = "ipet"
        self.config["background_pressure"] = 1012.0
        self.config["wind_conversion_factor"] = 0.93
        self.config["phi_spiral"] = 20.0
        self.config["asymmetry_option"] = "schwerdt1979"
        self.config["phi_trans"] = 45.0
        self.config["include_rainfall"] = True
        self.config["rho_air"] = 1.15  # used in the determination of parametric wind field
        self.config["spiderweb_radius"] = 400.0  # radius in km
        self.config["nr_radial_bins"] = 100
        self.config["nr_directional_bins"] = 36
        self.config["tref"] = "20000101 000000"

    def read_config(self, config_file):
        """Read the configuration file (toml format)""" 
        with open(config_file, "r") as f:
            cfg = toml.load(f)
        self.set_default_config()
        # Now loop over keys in cfg and update self.config
        for key in cfg:
            self.config[key] = cfg[key]

    def write_config(self, config_file):
        """Write the configuration file (toml format)"""        
        # Get path of the file
        path = os.path.dirname(config_file)
        if path and not os.path.exists(path):
            os.makedirs(path)
        with open(config_file, "w") as file:            
            toml.dump(self.config, file)

    def read_track(self, filename):
        # Read track file (automatically determine format)
        config = self.track.read(filename)
        if config is not None:
            # Old ddb_cyc format with config in it
            self.set_default_config()
            # Now loop over keys in cfg and update self.config
            for key in config:
                self.config[key] = config[key]

    def read_track_files_and_merge(self, track_file_list, tau=0):
        # Read track files and merge them
        self.track = TropicalCycloneTrack()
        for itrack, track_file in enumerate(track_file_list):
            track1 = TropicalCycloneTrack()
            if itrack == 0:
                # Do not use tau for the first track
                tau1 = 0
            else:
                tau1 = tau
            track1.read(track_file, tau=tau1)
            self.track.add(track1)

    # def resample_track(self, dt, method="spline"):
    #     """Resample the track to a new time step"""
    #     self.track.resample(dt, method=method)        

    def write_track(self, filename, **kwargs):
        # Write track file
        kwargs["name"] = self.name
        self.track.write(filename, **kwargs)

    def write_spiderweb(self, filename, **kwargs):
        # Write the spiderweb wind field to file
        self.spiderweb.write(filename, **kwargs) 

    def extend_track(self, **kwargs):
        """Extend the track"""
        self.track.extend(**kwargs)

    def shorten_track(self, **kwargs):
        """Shorten the track"""
        self.track.shorten(**kwargs)

    def add_track_point(self, datetime, **kwargs):
        """Add a track point"""
        self.track.add_point(datetime, **kwargs)    

    def compute_metric_track(self):
        """Make a copy of the track and convert units to metric"""
        self.track_metric = copy.deepcopy(self.track)
        self.track_metric.convert_units_imperial_to_metric(self.config)

    def compute_wind_field(self, filename=None, progress_bar=None, format="ascii", error_stats=False):
        """Compute the wind field using the specified wind profile"""

        # 1. First make a copy of self.track. Units in this copy will be converted to metric. Also missing values will be estimated.
        # 2. Convert units to km and m/s
        self.compute_metric_track()

        # 3. Compute forward speed
        self.track_metric.compute_forward_speed()

        # 4. Estimate missing values
        self.track_metric.estimate_missing_values(self.config) 

        # # 3. cut off track points with low wind speeds at beginning and end of track + extent track
        # self.cut_off_low_wind_speeds()
        # self.adjust_track_extent()

        # # 4. account for forward speed (computes several derivative values)
        # self.account_for_forward_speed()

        # 5. Initialize spiderweb grid
        self.spiderweb.initialize_grid(self.track,
                                       self.config["nr_radial_bins"],
                                       self.config["nr_directional_bins"],
                                       self.config["spiderweb_radius"] * 1000.0) # Creates the spiderweb arrays

        r = self.spiderweb.ds["range"].values / 1000 # in kilometres
        phi = self.spiderweb.ds["azimuth"].values

        # 6. Go over time steps in the track
        for it in range(len(self.track_metric.gdf)):
            print(it)

            # Progress bar
            if progress_bar:
                progress_bar.set_value(it)
                if progress_bar.was_canceled():
                    return

            # Copy radii data from vectors to matrix
            quadrants_speed = np.array([35.0, 50.0, 65.0, 100.0]) * knots_to_ms * self.config["wind_conversion_factor"]
            quadrants_radii = np.zeros((4, 4))
            quadrants_radii[0, 0] = self.track_metric.gdf.r35_ne[it]
            quadrants_radii[0, 1] = self.track_metric.gdf.r35_se[it]
            quadrants_radii[0, 2] = self.track_metric.gdf.r35_sw[it]
            quadrants_radii[0, 3] = self.track_metric.gdf.r35_nw[it]
            quadrants_radii[1, 0] = self.track_metric.gdf.r50_ne[it]
            quadrants_radii[1, 1] = self.track_metric.gdf.r50_se[it]
            quadrants_radii[1, 2] = self.track_metric.gdf.r50_sw[it]
            quadrants_radii[1, 3] = self.track_metric.gdf.r50_nw[it]
            quadrants_radii[2, 0] = self.track_metric.gdf.r65_ne[it]
            quadrants_radii[2, 1] = self.track_metric.gdf.r65_se[it]
            quadrants_radii[2, 2] = self.track_metric.gdf.r65_sw[it]
            quadrants_radii[2, 3] = self.track_metric.gdf.r65_nw[it]
            quadrants_radii[3, 0] = self.track_metric.gdf.r100_ne[it]
            quadrants_radii[3, 1] = self.track_metric.gdf.r100_se[it]
            quadrants_radii[3, 2] = self.track_metric.gdf.r100_sw[it]
            quadrants_radii[3, 3] = self.track_metric.gdf.r100_nw[it]

            # Get values ready
            dp = self.config["background_pressure"] - self.track_metric.gdf.pc[it]
            vmax = self.track_metric.gdf.vmax[it]
            pc = self.track_metric.gdf.pc[it]
            rmax = self.track_metric.gdf.rmw[it]
            pn = self.config["background_pressure"]
            coords = self.track_metric.gdf.geometry[it]
            latitude = coords.y

            # Get derivative values
            vt = np.sqrt(
                self.track_metric.gdf.vtx[it] ** 2 + self.track_metric.gdf.vty[it] ** 2
            )  # forward speed - magnitude
            phit = (
                np.arctan2(self.track_metric.gdf.vty[it], self.track_metric.gdf.vtx[it]) * 180 / np.pi
            )  # angle
            dpcdt = self.track_metric.gdf.dpcdt[it]  # pressure gradient in time

            # initialize temp arrays
            wind_speed = np.zeros((len(r), len(phi)))
            wind_to_direction_cart = np.zeros((len(r), len(phi)))
            wind_x = np.zeros((len(r), len(phi)))
            wind_y = np.zeros((len(r), len(phi)))
            air_pressure = np.zeros((len(r), len(phi)))
            precipitation = np.zeros((len(r), len(phi)))

            # We first determine vtcor and phicor. xn is set to 0.5. These may be overwritten if we have observations. By default we set phicor to 45 degrees.
            phia = self.config["phi_trans"] * np.pi / 180
            xn   = 0.5
            # phi = self.config["phicor"] * np.pi / 180
            if self.config["asymmetry_option"] == "schwerdt1979":
                # Schwerdt (1979)
                a = vt / knots_to_ms # convert back to kts
                vtcor = 1.5 * a**0.63 * knots_to_ms
            elif self.config["asymmetry_option"] == "mvo":
                vtcor = vt * 0.6
            elif self.config["asymmetry_option"] == "none":
                vtcor = 0.0
            else:
                raise Exception("This asymmetry_option is not supported")

            # Now check if fitting of the wind field is necessary
            # Holland et al. 2010
            if self.config["wind_profile"] == "holland2010":

                # Count the number of observations, so that we know if we need to do a fitting of with Holland (2010)
                fit_wind = False
                n = 0
                if vmax > 20:
                    for iq in range(np.size(quadrants_radii, 0)):
                        for irad in range(np.size(quadrants_radii, 1)):
                            if not np.isnan(quadrants_radii[iq, irad]):
                                n = n + 1
                if n > 0:
                    fit_wind = True

                # Do fitting of Holland 2010
                if fit_wind:
                    obs = {}
                    obs["quadrants_radii"] = quadrants_radii
                    obs["quadrants_speed"] = quadrants_speed
                    [xn, vtcor, phia] = fit_wind_field_holland2010(
                        vmax,
                        rmax,
                        pc,
                        vt,
                        phit,
                        pn,
                        self.config["phi_spiral"],
                        latitude,
                        dpcdt,
                        obs,
                    )

            if self.config["wind_profile"] == "holland2008":
                # Assume constant xn that follows a relationship described in 2008 paper
                xn = 0.6 * (1 - dp / 215)

            # Now compute the wind field (in goes vmax and vtcor). vr does not include asymmetry!
            if self.config["wind_profile"] == "holland1980" or self.config["wind_profile"] == "holland2010":
                vrel = vmax - vtcor
                [vr, pr] = holland2010(r, vrel, pc, pn, rmax, dpcdt, latitude, vt, xn)
            else:
                raise Exception("This wind_profile is not supported")

            # Loop over the directions
            for iphi in range(len(phi)):

                # Place wind speed
                wind_speed[:, iphi] = vr

                # Check which hemisphere we are
                if latitude >= 0:
                    # northern hemisphere
                    dr = 90.0 + phi[iphi] + self.config["phi_spiral"]
                else:
                    # southern hemisphere
                    dr = -90.0 + phi[iphi] - self.config["phi_spiral"]

                # Wind direction and pressure drop
                wind_to_direction_cart[:, iphi] = dr
                air_pressure[:, iphi] = 100 * pr

            # Wind speed with asymmetry
            # vnorm = wind_speed / np.max(wind_speed)
            vnorm = np.zeros(np.shape(wind_speed)) + 1.0
            u_prop = vtcor * np.cos((phit + phia) * np.pi / 180)
            v_prop = vtcor * np.sin((phit + phia) * np.pi / 180)
            wind_x = wind_speed * np.cos(wind_to_direction_cart * np.pi / 180) + u_prop * vnorm
            wind_y = wind_speed * np.sin(wind_to_direction_cart * np.pi / 180) + v_prop * vnorm

            # Rainfall (if required)
            if self.config["include_rainfall"] == True:
                # Add rainfall
                if self.config["rainfall_relationship"] == "ipet":
                    # IPET is a simple rainfall model relating pressure to rainfall rate
                    pdef = (self.config["background_pressure"] * 100 - air_pressure[0, 0]) / 100  # % hPa to Pa
                    pdef = max(pdef, 0.0)
                    for ip in range(len(r)):
                        if r[ip] < rmax:
                            precipitation[ip, :] = 1.14 + (0.12 * pdef)
                        else:
                            precipitation[ip, :] = (1.14 + (0.12 * pdef)) * np.exp(
                                -0.3 * ((r[ip] - rmax) / rmax)
                            )
                # More options to be added later
                # Bader, Bacla, etc.

            # Update data in spiderweb array
            self.spiderweb.ds["wind_x"].values[it, :, :] = wind_x
            self.spiderweb.ds["wind_y"].values[it, :, :] = wind_y
            self.spiderweb.ds["pressure"].values[it, :, :] = air_pressure
            if self.config["include_rainfall"]:
                self.spiderweb.ds["precipitation"].values[it, :, :] = precipitation


            # # Scale wind speeds back to ensure we always reach vmax
            # missing_factor = vmax / np.max(wind_speed)
            # wind_speed = wind_speed * missing_factor

        # Default is spiderweb in ascii
        if filename is not None:
            self.spiderweb.write(filename,
                                 format=format,
                                 background_pressure=self.config["background_pressure"],
                                 tref=self.config["tref"],
                                 include_rainfall=self.config["include_rainfall"])

        # If error stats are required
        if error_stats:
            # Compute the error stats
            error_stats = self.compute_track_error_stats()


    def get_track_from_spiderweb(self, replace=False):
        """Get the track from the spiderweb."""
        track = self.spiderweb.get_track(self.config)
        if replace:
            self.track = track
        return track    

    def compute_track_error_stats(self):
        """Compute error statistics"""
        resulting_track = self.spiderweb.get_track(self.config)
        r_best   = np.empty((len(resulting_track.gdf), 4, 4)) * np.nan
        r_result = np.empty((len(resulting_track.gdf), 4, 4)) * np.nan
        r_diff   = np.empty((len(resulting_track.gdf), 4, 4)) * np.nan
        r_diff_rel  = np.empty((len(resulting_track.gdf), 4, 4)) * np.nan
        r_diff_radius_mean = np.zeros((len(resulting_track.gdf), 4))
        r_diff_radius_rel_mean = np.zeros((len(resulting_track.gdf), 4))
        r_diff_mean = np.zeros((len(resulting_track.gdf)))
        r_diff_rel_mean = np.zeros((len(resulting_track.gdf)))
        rstr = ["r35", "r50", "r65", "r100"]
        vstr = ["ne", "se", "sw", "nw"]
        # Loop through time
        for it in range(len(resulting_track.gdf)):
            print("Time step : " + str(it))
            for ir in range(4):
                iok = 0
                ibst = 0
                ires = 0
                for iv in range(4):
                    r_best[it, ir, iv]   = self.track.gdf[rstr[ir] + "_" + vstr[iv]][it]
                    r_result[it, ir, iv] = resulting_track.gdf[rstr[ir] + "_" + vstr[iv]][it]
                    # Compute the error statistics for this wind speed
                    r_diff[it, ir, iv] = r_result[it, ir, iv] - r_best[it, ir, iv]
                    r_diff_rel[it, ir, iv] = 100 * (r_result[it, ir, iv] - r_best[it, ir, iv]) / r_best[it, ir, iv]
                    if not np.isnan(r_best[it, ir, iv]) and not np.isnan(r_result[it, ir, iv]):
                        iok += 1
                    if not np.isnan(r_best[it, ir, iv]) and np.isnan(r_result[it, ir, iv]):
                        ibst += 1
                    if np.isnan(r_best[it, ir, iv]) and not np.isnan(r_result[it, ir, iv]):
                        ires += 1    
                r_diff_radius_mean[it, ir] = np.nanmean(r_diff[it, ir, :])
                r_diff_radius_rel_mean[it, ir] = np.nanmean(r_diff_rel[it, ir, :])

#                print("Mean error for " + rstr[ir] + " : " + '{0:.1f}'.format(r_diff_radius_mean[it, ir]) + " NM (" + '{0:.0f}'.format(r_diff_radius_rel_mean[it, ir]) + " %)" + "    - " + str(iok) + " OK, " + str(ibst) + " Best, " + str(ires) + " Result")
                print("Mean error for " + rstr[ir] + " : " + '{0:.1f}'.format(r_diff_radius_mean[it, ir]) + " NM (" + '{0:.0f}'.format(r_diff_radius_rel_mean[it, ir]) + " %)")

            r_diff_mean[it] = np.nanmean(r_diff_radius_mean[it, :])
            r_diff_rel_mean[it] = np.nanmean(r_diff_radius_rel_mean[it, :])

            print("Mean error for all : " + '{0:.1f}'.format(r_diff_mean[it]) + " NM (" + '{0:.0f}'.format(r_diff_rel_mean[it]) + " %)")

    def make_ensemble(self, **kwargs):
        """Make ensemble"""
        ensemble = TropicalCycloneEnsemble(self, **kwargs)        
        ensemble.generate()
        return ensemble
    
    def merge_with_meteo_dataset(self):
        """Merge with meteo dataset. Not implemented yet."""
        pass

    def get_track_from_meteo_dataset(self, meteo_dataset):
        """Get the track from a meteo dataset. Not implemented yet."""
        # Find the track(s)
        pass

    def get_wind_field_from_meteo_dataset(self, meteo_dataset, filename=None, format="ascii"):
        """Get the wind field from a meteo dataset"""

        # Track has already been defined
        self.spiderweb.initialize_grid(self.track,
                                       self.config["nr_radial_bins"],
                                       self.config["nr_directional_bins"],
                                       self.config["spiderweb_radius"] * 1000) # Creates the spiderweb arrays

        # Loop over time steps in the track
        for it in range(len(self.track.gdf)):
            # Get spiderweb grid coordinates
            tt = datetime.strptime(self.track.gdf.datetime[it], dateformat_module)
            xx = self.spiderweb.ds["lon"].values[it, :, :]
            yy = self.spiderweb.ds["lat"].values[it, :, :]
            # Get the data at the specified time (really should start doing this with xarray!)
            self.spiderweb.ds["wind_x"].values[it, :, :] = meteo_dataset.interpolate_variable("wind_u", tt, xx, yy)
            self.spiderweb.ds["wind_y"].values[it, :, :] = meteo_dataset.interpolate_variable("wind_v", tt, xx, yy)
            self.spiderweb.ds["pressure"].values[it, :, :] = meteo_dataset.interpolate_variable("barometric_pressure", tt, xx, yy)
            self.spiderweb.ds["precipitation"].values[it, :, :] = meteo_dataset.interpolate_variable("precipitation", tt, xx, yy)

        # Default is spiderweb in ascii
        if filename is not None:
            self.spiderweb.write(filename,
                                 format=format,
                                 background_pressure=self.config["background_pressure"],
                                 tref=self.config["tref"],
                                 include_rainfall=self.config["include_rainfall"])


    def get_wind_field_from_coamps(self, coamps_path, coamps_cycle, filename=None, format="ascii"):
        """Get the wind field from coamps-tc"""

        import xarray as xr
        from cht_utils.misc_tools import interp2

        # # Track has already been defined
        self.spiderweb.initialize_grid(self.track,
                                       self.config["nr_radial_bins"],
                                       self.config["nr_directional_bins"],
                                       self.config["spiderweb_radius"] * 1000) # Creates the spiderweb arrays

        # 6. Loop over time steps in the track
        for it in range(len(self.track.gdf)):

            # Get time
            tt = datetime.strptime(self.track.gdf.datetime[it], dateformat_module)

            # Get spiderweb grid coordinates
            xx = self.spiderweb.ds["lon"].values[it, :, :]
            yy = self.spiderweb.ds["lat"].values[it, :, :]

            # Convert coamps_cycle to time
            coamps_cycle_time = datetime.strptime(coamps_cycle, "%Y%m%d%H")

            # Get the time difference between tt and coamps_cycle_time in hours
            tau = int(((tt - coamps_cycle_time).total_seconds() / 3600))
            taustr = "tau" + str(tau).zfill(3)

            print("Time step : " + str(it) + " - " + str(tt) + " - " + taustr)

            # First do high-res d03

            coamps_file = os.path.join(coamps_path, "coamps-tc_d03_" + coamps_cycle + "_" + taustr + ".nc")

            if not os.path.exists(coamps_file):
                print("File does not exist : " + coamps_file)
                continue
                                       
            ds = xr.open_dataset(coamps_file)
            # ds = ds.sel(lon=slice(xx.min(), xx.max()), lat=slice(yy.min(), yy.max()))

            lon = ds["lon"].values
            lat = ds["lat"].values
            uu = ds["uuwind"].values
            vv = ds["vvwind"].values
            pp = ds["slpres"].values * 100
            rr = ds["precip"].values

            lon = lon[0, :] - 360.0
            lat = lat[:, 0]

            # Get the data at the specified time (really should start doing this with xarray!)
            self.spiderweb.ds["wind_x"].values[it, :, :] = interp2(lon, lat, uu, xx, yy)
            self.spiderweb.ds["wind_y"].values[it, :, :] = interp2(lon, lat, vv, xx, yy)
            self.spiderweb.ds["pressure"].values[it, :, :] = interp2(lon, lat, pp, xx, yy)
            self.spiderweb.ds["precipitation"].values[it, :, :] = interp2(lon, lat, rr, xx, yy)

            # Now do low-res d02
            # But only where we have missing values

            coamps_file = os.path.join(coamps_path, "coamps-tc_d02_" + coamps_cycle + "_" + taustr + ".nc")

            if not os.path.exists(coamps_file):
                print("File does not exist : " + coamps_file)
                continue

            ds = xr.open_dataset(coamps_file)

            lon = ds["lon"].values
            lat = ds["lat"].values
            uu = ds["uuwind"].values
            vv = ds["vvwind"].values
            pp = ds["slpres"].values * 100
            rr = ds["precip"].values

            lon = lon[0, :] - 360.0
            lat = lat[:, 0]

            # Get the data at the specified time (really should start doing this with xarray!)
            self.spiderweb.ds["wind_x"].values[it, :, :][np.isnan(self.spiderweb.ds["wind_x"].values[it, :, :])] = interp2(lon, lat, uu, xx, yy)[np.isnan(self.spiderweb.ds["wind_x"].values[it, :, :])]
            self.spiderweb.ds["wind_y"].values[it, :, :][np.isnan(self.spiderweb.ds["wind_y"].values[it, :, :])] = interp2(lon, lat, vv, xx, yy)[np.isnan(self.spiderweb.ds["wind_y"].values[it, :, :])]
            self.spiderweb.ds["pressure"].values[it, :, :][np.isnan(self.spiderweb.ds["pressure"].values[it, :, :])] = interp2(lon, lat, pp, xx, yy)[np.isnan(self.spiderweb.ds["pressure"].values[it, :, :])]
            self.spiderweb.ds["precipitation"].values[it, :, :][np.isnan(self.spiderweb.ds["precipitation"].values[it, :, :])] = interp2(lon, lat, rr, xx, yy)[np.isnan(self.spiderweb.ds["precipitation"].values[it, :, :])]

        # Replace nan with 0.0 in wind field
        self.spiderweb.ds["wind_x"].values[np.isnan(self.spiderweb.ds["wind_x"].values)] = 0.0
        self.spiderweb.ds["wind_y"].values[np.isnan(self.spiderweb.ds["wind_y"].values)] = 0.0  
        # Replace nan with background pressure        
        self.spiderweb.ds["pressure"].values[np.isnan(self.spiderweb.ds["pressure"].values)] = self.config["background_pressure"]
        # Replace nan with 0.0 in precipitation
        self.spiderweb.ds["precipitation"].values[np.isnan(self.spiderweb.ds["precipitation"].values)] = 0.0

        # Default is spiderweb in ascii
        if filename is not None:
            self.spiderweb.write(filename,
                                 format=format,
                                 background_pressure=self.config["background_pressure"],
                                 tref=self.config["tref"],
                                 include_rainfall=self.config["include_rainfall"])


    def to_gdf(self, filename=None):
        """Convert track to GeoDataFrame, and optionally write to file."""
        return self.track.to_gdf(filename=filename)


# Definition to find if this is a land point
def analyze_points_with_shapefile(
    shapefile_polygon, shapefile_polyline, latitudes, longitudes
):
    from shapely.geometry import Point
    from shapely.ops import nearest_points

    # Read the shapefile
    shapefile_polygon = gpd.read_file(shapefile_polygon)
    shapefile_polyline = gpd.read_file(shapefile_polyline)

    # Empty results
    results = []

    for latitude, longitude in zip(latitudes, longitudes):
        # Create a Point object from the latitude and longitude
        point = Point(longitude, latitude)

        # Check if the point falls within any polygon of the shapefile
        is_within = shapefile_polygon.contains(point).any()

        if is_within:
            # Find the nearest vertex from the point
            nearest_vertex = nearest_points(point, shapefile_polyline.unary_union)[1]

            # Calculate the distance between the point and the nearest vertex
            distance_km = point.distance(nearest_vertex) * 111.32

            result = {
                "latitude": latitude,
                "longitude": longitude,
                "is_within": True,
                "distance_km": distance_km,
            }
        else:
            result = {
                "latitude": latitude,
                "longitude": longitude,
                "is_within": False,
                "distance_km": None,
            }

        results.append(result)

    return results
